import os
import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

import wisdem.rotorse.rail_transport as rt


class TestRT(unittest.TestCase):
    def testRailTransport(self):
        prob = om.Problem(reports=False)

        # Data structure generated with
        # np.savez('rail.npz', blade_ref_axis = inputs['blade_ref_axis'],
        # chord = inputs['chord'], theta = inputs['theta'],
        # x_sc = inputs['x_sc'], y_sc = inputs['y_sc'],
        # A = inputs['A'], rhoA = inputs['rhoA'],
        # rhoJ = inputs['rhoJ'], GJ = inputs['GJ'],
        # EA = inputs['EA'], EIxx = inputs['EIxx'],
        # EIyy = inputs['EIyy'], EIxy = inputs['EIxy'],
        # coord_xy_interp = inputs['coord_xy_interp'])

        path2npz = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "rail.npz"
        data = np.load(path2npz)

        blade_ref_axis = data["blade_ref_axis"]
        chord = data["chord"]
        theta = data["theta"]
        x_sc = data["x_sc"]
        y_sc = data["y_sc"]
        A = data["A"]
        rhoA = data["rhoA"]
        rhoJ = data["rhoJ"]
        GJ = data["GJ"]
        EA = data["EA"]
        EIxx = data["EIxx"]
        EIyy = data["EIyy"]
        EIxy = data["EIxy"]
        coord_xy_interp = data["coord_xy_interp"]

        n_span = np.size(coord_xy_interp, axis=0)
        n_xy = np.size(coord_xy_interp, axis=1)
        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["RotorSE"] = {}
        modeling_options["WISDEM"]["RotorSE"]["n_span"] = n_span
        modeling_options["WISDEM"]["RotorSE"]["n_xy"] = n_xy

        prob.model.add_subsystem("powercurve", rt.RailTransport(modeling_options=modeling_options), promotes=["*"])

        prob.setup()

        prob.set_val("blade_ref_axis", blade_ref_axis, units="m")
        prob.set_val("chord", chord, units="m")
        prob.set_val("theta", theta, units="deg")
        prob.set_val("x_sc", x_sc, units="m")
        prob.set_val("y_sc", y_sc, units="m")
        prob.set_val("A", A, units="m**2")
        prob.set_val("rhoA", rhoA, units="kg/m")
        prob.set_val("rhoJ", rhoJ, units="kg*m")
        prob.set_val("GJ", GJ, units="N*m**2")
        prob.set_val("EA", EA, units="N")
        prob.set_val("EIxx", EIxx, units="N*m**2")
        prob.set_val("EIyy", EIyy, units="N*m**2")
        prob.set_val("EIxy", EIxy, units="N*m**2")
        prob.set_val("coord_xy_interp", coord_xy_interp)

        prob.run_model()

        print(prob["constr_LV_4axle_horiz"][0])
        print(prob["constr_LV_4axle_horiz"][1])
        print(prob["constr_LV_8axle_horiz"][0])
        print(prob["constr_LV_8axle_horiz"][1])

        self.assertAlmostEqual(prob["constr_LV_4axle_horiz"][0], 2.205225587616127, places=1)
        self.assertAlmostEqual(prob["constr_LV_4axle_horiz"][1], 2.5000819086864334, places=1)
        # self.assertAlmostEqual(prob["constr_LV_8axle_horiz"][0], 0.8586925171861572, places=2)
        # self.assertAlmostEqual(prob["constr_LV_8axle_horiz"][1], 1.008140486233143, places=2)
        # self.assertAlmostEqual(max(abs(prob["constr_strainPS"])), 0.39050013815507656, places=2)
        # self.assertAlmostEqual(max(abs(prob["constr_strainSS"])), 0.432439251608084, places=2)


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestRT),
    ]
    return unittest.TestSuite(suite)


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



"""Tests for the `MonopileDesign` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy
from itertools import product

import pytest

from wisdem.orbit.phases.design import MonopileDesign

base = {
    "site": {"depth": 50, "mean_windspeed": 9},
    "plant": {"num_turbines": 20},
    "turbine": {
        "rotor_diameter": 150,
        "hub_height": 110,
        "rated_windspeed": 11,
    },
}

turbines = [
    {"rotor_diameter": 155, "hub_height": 100, "rated_windspeed": 12},
    {"rotor_diameter": 180, "hub_height": 112, "rated_windspeed": 12},
    {"rotor_diameter": 205, "hub_height": 125, "rated_windspeed": 12},
    {"rotor_diameter": 222, "hub_height": 136, "rated_windspeed": 12},
    {"rotor_diameter": 248, "hub_height": 149, "rated_windspeed": 12},
]


@pytest.mark.parametrize(
    "depth,mean_ws,turbine",
    product(range(10, 51, 10), range(8, 13, 1), turbines),
)
def test_paramater_sweep(depth, mean_ws, turbine):
    config = {
        "site": {"depth": depth, "mean_windspeed": mean_ws},
        "plant": {"num_turbines": 20},
        "turbine": turbine,
    }

    m = MonopileDesign(config)
    m.run()

    # Check valid monopile length
    assert 10 < m._outputs["monopile"]["length"] < 130

    # Check valid monopile diameter
    assert 4 < m._outputs["monopile"]["diameter"] < 13

    # Check valid monopile mass
    assert 200 < m._outputs["monopile"]["mass"] < 2500

    # Check valid transition piece diameter
    assert 4 < m._outputs["transition_piece"]["diameter"] < 14


def test_monopile_kwargs():
    test_kwargs = {
        "yield_stress": 400000000,
        "load_factor": 1.25,
        "material_factor": 1.2,
        "monopile_density": 9000,
        "monopile_modulus": 220e9,
        "soil_coefficient": 4500000,
        "air_density": 1.125,
        "weibull_scale_factor": 10,
        "weibull_shape_factor": 1.87,
        "turb_length_scale": 350,
    }

    m = MonopileDesign(base)
    m.run()
    base_results = m._outputs["monopile"]

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["monopile_design"] = {}
        config["monopile_design"][k] = v

        m = MonopileDesign(config)
        m.run()
        results = m._outputs["monopile"]

        assert results != base_results


def test_transition_piece_kwargs():
    test_kwargs = {
        # Transition piece specific
        "monopile_tp_connection_thickness": 0.005,
        "transition_piece_density": 8200,
        "transition_piece_thickness": 0.12,
        "transition_piece_length": 30,
    }

    m = MonopileDesign(base)
    m.run()
    base_results = m._outputs["transition_piece"]

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["monopile_design"] = {}
        config["monopile_design"][k] = v

        m = MonopileDesign(config)
        m.run()
        results = m._outputs["transition_piece"]

        assert results != base_results



# 0 ---------- (marker for docs)
import numpy as np
import openmdao.api as om

from wisdem.nrelcsm.nrel_csm_orig import aep_csm
from wisdem.nrelcsm.nrel_csm_mass_2015 import nrel_csm_2015

# 0 ---------- (marker for docs)

# 1 ---------- (marker for docs)
# OpenMDAO Problem instance
prob = om.Problem(reports=False)
prob.model = nrel_csm_2015()
prob.setup()

# Initialize AEP calculator from CSM model
aep_instance = aep_csm()
# 1 ---------- (marker for docs)

# 2 ---------- (marker for docs)
# Initialize variables for NREL CSM
prob["turbine_class"] = -1  # Sets blade mass based on user input, not auto-determined
prob["blade_number"] = 3
prob["blade_has_carbon"] = False
prob["max_tip_speed"] = max_tip_speed = 90.0
prob["max_efficiency"] = max_efficiency = 0.9
prob["main_bearing_number"] = 2
prob["crane"] = True
# 2 ---------- (marker for docs)

# 3 ---------- (marker for docs)
# Initialize variables for AEP calculation
opt_tsr = 9.0  # Optimal tip speed ratio
max_Cp = 0.47  # Max (aerodynamic) power coefficient
max_Ct = 0.8  # Max thrust coefficient
max_eff = 0.95  # Drivetrain efficiency
cut_in = 4.0  # m/s
cut_out = 25.0  # m/s
altitude = 0.0  # m (assume sea level)
rho_air = 1.225  # kg/m^3
array_losses = 0.0  # Focusing on single turbine
availability = 1.0  # Assume 100% uptime
soiling_losses = 0.0  # Ignore this for now
turbine_number = 1  # Focus on single turbine
weibull_k = 2.0  # Weibull shape parameter
# 3 ---------- (marker for docs)

# 4 ---------- (marker for docs)
# Set Design of Experiment (DoE) parametric sweep bounds
machine_rating = 1e3 * np.arange(2.0, 10.1, 1.0)  # kW
rotor_diameter = np.arange(60, 201.0, 20.0)  # m
blade_mass_exp = np.arange(2.1, 2.41, 0.1)  # Relationship between blade length and mass
shear_exp = np.arange(0.1, 0.31, 0.1)
wind_speed = np.arange(5.0, 11.1, 1.0)  # m/s
# 4 ---------- (marker for docs)

# 5 ---------- (marker for docs)
# Enumerate DoE through tensor multiplication and then flatten to get vector of all of the runs
[Rating, Diameter, Bladeexp, Shear, WindV] = np.meshgrid(
    machine_rating, rotor_diameter, blade_mass_exp, shear_exp, wind_speed
)

# Shift to flattened arrays to run through each scenario easily
Rating = Rating.flatten()
Diameter = Diameter.flatten()
Bladeexp = Bladeexp.flatten()
Shear = Shear.flatten()
WindV = WindV.flatten()

# Initialize output containers
tcc = np.zeros(Rating.shape)
aep = np.zeros(Rating.shape)
# 5 ---------- (marker for docs)

# 6 ---------- (marker for docs)
# Calculation loop
npts = Rating.size
print("Running, ", npts, " points in the parametric study")
for k in range(npts):
    # Populate remaining NREL CSM inputs for this iteration
    prob["machine_rating"] = Rating[k]
    prob["rotor_diameter"] = Diameter[k]
    prob["blade_user_exp"] = Bladeexp[k]
    prob["tower_length"] = hub_height = 0.5 * Diameter[k] + 30.0

    # Compute Turbine capital cost using the NREL CSM (2015) and store the result
    prob.run_model()
    tcc[k] = float(prob["turbine_cost_kW"])

    # Compute AEP using original CSM function and store result
    aep_instance.compute(
        Rating[k],
        max_tip_speed,
        Diameter[k],
        max_Cp,
        opt_tsr,
        cut_in,
        cut_out,
        hub_height,
        altitude,
        rho_air,
        max_efficiency,
        max_Ct,
        soiling_losses,
        array_losses,
        availability,
        turbine_number,
        Shear[k],
        WindV[k],
        weibull_k,
    )
    aep[k] = aep_instance.aep.net_aep

    # Progress update to screen every 1000 iterations
    if np.mod(k, 1000) == 0:
        print("...Completed iteration, ", k)
# 6 ---------- (marker for docs)

# 7 ---------- (marker for docs)
# Write outputs to csv file for later post-processing
alldata = np.c_[Rating, Diameter, Bladeexp, Shear, WindV, tcc, aep]
header = "Rating [kW],Rotor Diam [m],Blade Mass Exp,Shear Exp,Wind Vel [m/s],TCC [USD/kW],AEP [kWh/yr]"
np.savetxt("parametric_scaling.csv", alldata, delimiter=",", header=header)
# 7 ---------- (marker for docs)



import numpy as np
import openmdao.api as om

import wisdem.commonse.utilities as util
from wisdem.commonse import gravity
from wisdem.commonse.cylinder_member import NULL, MEMMAX

NNODES_MAX = 500  # 1000
NELEM_MAX = 500  # 1000
RIGID = 1e30
EPS = 1e-6


class PlatformFrame(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]
        self.shape = opt["floating"]["members"]["outer_shape"]

        for k in range(n_member):
            self.add_input(f"member{k}:nodes_xyz", NULL * np.ones((MEMMAX, 3)), units="m")
            self.add_input(f"member{k}:nodes_r", NULL * np.ones(MEMMAX), units="m")
            if self.shape[k] == "circular":
                self.add_input(f"member{k}:section_D", NULL * np.ones(MEMMAX), units="m")
            elif self.shape[k] == "rectangular":
                self.add_input(f"member{k}:section_a", NULL * np.ones(MEMMAX), units="m")
                self.add_input(f"member{k}:section_b", NULL * np.ones(MEMMAX), units="m")
            self.add_input(f"member{k}:Iwaterx", 0.0, units="m**4")
            self.add_input(f"member{k}:Iwatery", 0.0, units="m**4")
            self.add_input(f"member{k}:section_t", NULL * np.ones(MEMMAX), units="m")
            self.add_input(f"member{k}:section_A", NULL * np.ones(MEMMAX), units="m**2")
            self.add_input(f"member{k}:section_Asx", NULL * np.ones(MEMMAX), units="m**2")
            self.add_input(f"member{k}:section_Asy", NULL * np.ones(MEMMAX), units="m**2")
            self.add_input(f"member{k}:section_Ixx", NULL * np.ones(MEMMAX), units="kg*m**2")
            self.add_input(f"member{k}:section_Iyy", NULL * np.ones(MEMMAX), units="kg*m**2")
            self.add_input(f"member{k}:section_J0", NULL * np.ones(MEMMAX), units="kg*m**2")
            self.add_input(f"member{k}:section_rho", NULL * np.ones(MEMMAX), units="kg/m**3")
            self.add_input(f"member{k}:section_E", NULL * np.ones(MEMMAX), units="Pa")
            self.add_input(f"member{k}:section_G", NULL * np.ones(MEMMAX), units="Pa")
            self.add_input(f"member{k}:section_TorsC", NULL * np.ones(MEMMAX), units="m**3")
            self.add_input(f"member{k}:section_sigma_y", NULL * np.ones(MEMMAX), units="Pa")
            self.add_input(f"member{k}:idx_cb", 0)
            self.add_input(f"member{k}:buoyancy_force", 0.0, units="N")
            self.add_input(f"member{k}:displacement", 0.0, units="m**3")
            self.add_input(f"member{k}:center_of_buoyancy", np.zeros(3), units="m")
            self.add_input(f"member{k}:center_of_mass", np.zeros(3), units="m")
            self.add_input(f"member{k}:ballast_mass", 0.0, units="kg")
            self.add_input(f"member{k}:total_mass", 0.0, units="kg")
            self.add_input(f"member{k}:total_cost", 0.0, units="USD")
            self.add_input(f"member{k}:I_total", np.zeros(6), units="kg*m**2")
            self.add_input(f"member{k}:Awater", 0.0, units="m**2")
            self.add_input(f"member{k}:added_mass", np.zeros(6), units="kg")
            self.add_input(f"member{k}:waterline_centroid", np.zeros(2), units="m")
            self.add_input(f"member{k}:variable_ballast_capacity", val=0.0, units="m**3")

        self.add_input("transition_node", np.zeros(3), units="m")
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_cost", 0.0, units="USD")

        self.add_output("transition_piece_I", np.zeros(6), units="kg*m**2")

        self.add_output("platform_nodes", NULL * np.ones((NNODES_MAX, 3)), units="m")
        self.add_output("platform_Fnode", NULL * np.ones((NNODES_MAX, 3)), units="N")
        self.add_output("platform_Rnode", NULL * np.ones(NNODES_MAX), units="m")
        self.add_output("platform_elem_n1", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_output("platform_elem_n2", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_output("platform_elem_L", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("platform_elem_D", NULL * np.ones(NELEM_MAX), units="m") # elem_D, a, b are added for both circular and rectangular
        self.add_output("platform_elem_a", NULL * np.ones(NELEM_MAX), units="m") # zeros when not the corresponding type
        self.add_output("platform_elem_b", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("platform_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("platform_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("platform_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("platform_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("platform_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("platform_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("platform_elem_J0", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("platform_elem_rho", NULL * np.ones(NELEM_MAX), units="kg/m**3")
        self.add_output("platform_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("platform_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("platform_elem_TorsC", NULL * np.ones(NELEM_MAX), units="m**3")
        self.add_output("platform_elem_sigma_y", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_discrete_output("platform_elem_memid", [-1] * NELEM_MAX)
        self.add_output("platform_displacement", 0.0, units="m**3")
        self.add_output("platform_center_of_buoyancy", np.zeros(3), units="m")
        self.add_output("platform_hull_center_of_mass", np.zeros(3), units="m")
        self.add_output("platform_centroid", np.zeros(3), units="m")
        self.add_output("platform_ballast_mass", 0.0, units="kg")
        self.add_output("platform_hull_mass", 0.0, units="kg")
        self.add_output("platform_I_hull", np.zeros(6), units="kg*m**2")
        self.add_output("platform_cost", 0.0, units="USD")
        self.add_output("platform_Awater", 0.0, units="m**2")
        # self.add_output("platform_Iwater", 0.0, units="m**4")
        self.add_output("platform_Iwaterx", 0.0, units="m**4")
        self.add_output("platform_Iwatery", 0.0, units="m**4")
        self.add_output("platform_added_mass", np.zeros(6), units="kg")
        self.add_output("platform_variable_capacity", np.zeros(n_member), units="m**3")

        self.node_mem2glob = {}

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Seems like we have to run this each time as numbering can change during optimization
        self.node_mem2glob = {}
        self.set_connectivity(inputs, outputs)

        self.set_node_props(inputs, outputs)
        self.set_element_props(inputs, outputs, discrete_inputs, discrete_outputs)

    def set_connectivity(self, inputs, outputs):
        # Load in number of members
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]

        # Initialize running lists across all members
        nodes_temp = np.empty((0, 3))
        elem_n1 = np.array([], dtype=np.int_)
        elem_n2 = np.array([], dtype=np.int_)

        # Look over members and grab all nodes and internal connections
        for k in range(n_member):
            inode_xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(inode_xyz[:, 0] == NULL)[0][0]
            inode_xyz = inode_xyz[:inodes, :]
            inode_range = np.arange(inodes - 1)

            n = nodes_temp.shape[0]
            for ii in range(inodes):
                self.node_mem2glob[(k, ii)] = n + ii

            elem_n1 = np.append(elem_n1, n + inode_range)
            elem_n2 = np.append(elem_n2, n + inode_range + 1)
            nodes_temp = np.append(nodes_temp, inode_xyz, axis=0)

        # Reveal connectivity by using mapping to unique node positions
        nodes, idx, inv = np.unique(nodes_temp.round(8), axis=0, return_index=True, return_inverse=True)
        nnode = nodes.shape[0]
        outputs["platform_nodes"] = NULL * np.ones((NNODES_MAX, 3))
        outputs["platform_nodes"][:nnode, :] = nodes
        outputs["platform_centroid"] = nodes.mean(axis=0)

        # Use mapping to set references to node joints
        nelem = elem_n1.size
        outputs["platform_elem_n1"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["platform_elem_n2"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["platform_elem_L"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["platform_elem_n1"][:nelem] = idx1 = np.int_(inv[elem_n1])
        outputs["platform_elem_n2"][:nelem] = idx2 = np.int_(inv[elem_n2])
        outputs["platform_elem_L"][:nelem] = np.sqrt(np.sum((nodes[idx2, :] - nodes[idx1, :]) ** 2, axis=1))

        # Update global 2 member mappings
        for k in self.node_mem2glob.keys():
            self.node_mem2glob[k] = inv[self.node_mem2glob[k]]

    def set_node_props(self, inputs, outputs):
        # Load in number of members
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]

        # Number of valid nodes
        node_platform = outputs["platform_nodes"]
        nnode = np.where(node_platform[:, 0] == NULL)[0][0]
        node_platform = node_platform[:nnode, :]

        # Find greatest radius of all members at node intersections
        Rnode = np.zeros(nnode)
        for k in range(n_member):
            irnode = inputs[f"member{k}:nodes_r"]
            n = np.where(irnode == NULL)[0][0]
            for ii in range(n):
                iglob = self.node_mem2glob[(k, ii)]
                Rnode[iglob] = np.array([Rnode[iglob], irnode[ii]]).max()

        # Find forces on nodes
        Fnode = np.zeros((nnode, 3))
        for k in range(n_member):
            icb = int(inputs[f"member{k}:idx_cb"][0])
            iglob = self.node_mem2glob[(k, icb)]
            Fnode[iglob, 2] += inputs[f"member{k}:buoyancy_force"][0]

        # Get transition piece inertial properties
        itrans_platform = util.closest_node(node_platform, inputs["transition_node"])
        m_trans = float(inputs["transition_piece_mass"][0])
        r_trans = Rnode[itrans_platform]
        I_trans = m_trans * r_trans**2.0 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]
        outputs["transition_piece_I"] = I_trans

        # Store outputs
        outputs["platform_Rnode"] = NULL * np.ones(NNODES_MAX)
        outputs["platform_Rnode"][:nnode] = Rnode
        outputs["platform_Fnode"] = NULL * np.ones((NNODES_MAX, 3))
        outputs["platform_Fnode"][:nnode, :] = Fnode

    def set_element_props(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Load in number of members
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]
        self.shape = opt["floating"]["members"]["outer_shape"]

        # Initialize running lists across all members
        elem_D = np.array([])
        elem_a = np.array([])
        elem_b = np.array([])
        elem_t = np.array([])
        elem_A = np.array([])
        elem_Asx = np.array([])
        elem_Asy = np.array([])
        elem_Ixx = np.array([])
        elem_Iyy = np.array([])
        elem_J0 = np.array([])
        elem_rho = np.array([])
        elem_E = np.array([])
        elem_G = np.array([])
        elem_TorsC = np.array([])
        elem_sigy = np.array([])
        elem_memid = np.array([], dtype=np.int_)

        mass = 0.0
        m_ball = 0.0
        cost = 0.0
        volume = 0.0
        Awater = 0.0
        # Iwater = 0.0
        Iwaterx = 0.0
        Iwatery = 0.0
        cg_plat = np.zeros(3)
        cb_plat = np.zeros(3)
        centroid = outputs["platform_centroid"][:2]
        variable_capacity = np.zeros(n_member)

        # Append all member data
        for k in range(n_member):
            n = np.where(inputs[f"member{k}:section_A"] == NULL)[0][0]
            if self.shape[k] == "circular":
                elem_D = np.append(elem_D, inputs[f"member{k}:section_D"][:n])
                elem_a = np.append(elem_a, np.zeros(n))
                elem_b = np.append(elem_b, np.zeros(n))
            elif self.shape[k] == "rectangular":
                elem_D = np.append(elem_D, np.zeros(n))
                elem_a = np.append(elem_a, inputs[f"member{k}:section_a"][:n])
                elem_b = np.append(elem_b, inputs[f"member{k}:section_b"][:n])
            elem_t = np.append(elem_t, inputs[f"member{k}:section_t"][:n])
            elem_A = np.append(elem_A, inputs[f"member{k}:section_A"][:n])
            elem_Asx = np.append(elem_Asx, inputs[f"member{k}:section_Asx"][:n])
            elem_Asy = np.append(elem_Asy, inputs[f"member{k}:section_Asy"][:n])
            elem_Ixx = np.append(elem_Ixx, inputs[f"member{k}:section_Ixx"][:n])
            elem_Iyy = np.append(elem_Iyy, inputs[f"member{k}:section_Iyy"][:n])
            elem_J0 = np.append(elem_J0, inputs[f"member{k}:section_J0"][:n])
            elem_rho = np.append(elem_rho, inputs[f"member{k}:section_rho"][:n])
            elem_E = np.append(elem_E, inputs[f"member{k}:section_E"][:n])
            elem_G = np.append(elem_G, inputs[f"member{k}:section_G"][:n])
            elem_TorsC = np.append(elem_TorsC, inputs[f"member{k}:section_TorsC"][:n])
            elem_sigy = np.append(elem_sigy, inputs[f"member{k}:section_sigma_y"][:n])
            elem_memid = np.append(elem_memid, k * np.ones(n, dtype=np.int_))

            # Mass, volume, cost tallies
            imass = inputs[f"member{k}:total_mass"]
            ivol = inputs[f"member{k}:displacement"]

            mass += imass
            volume += ivol
            cost += inputs[f"member{k}:total_cost"]
            m_ball += inputs[f"member{k}:ballast_mass"]
            Awater_k = inputs[f"member{k}:Awater"]
            Awater += Awater_k
            # y-coordinate for roll metacentric height
            # x-coordinate for pitch meracentric height
            Rwaterx = inputs[f"member{k}:waterline_centroid"][1] - centroid[1]
            Rwatery = inputs[f"member{k}:waterline_centroid"][0] - centroid[0]
            Iwaterx += inputs[f"member{k}:Iwaterx"] + Awater_k * Rwaterx**2
            Iwatery += inputs[f"member{k}:Iwatery"] + Awater_k * Rwatery**2
            variable_capacity[k] = inputs[f"member{k}:variable_ballast_capacity"][0]

            # Center of mass / buoyancy tallies
            cg_plat += imass * inputs[f"member{k}:center_of_mass"]
            cb_plat += ivol * inputs[f"member{k}:center_of_buoyancy"]

        # Add transition piece
        m_trans = inputs["transition_piece_mass"]
        cg_trans = inputs["transition_node"]
        I_trans = util.assembleI(outputs["transition_piece_I"])
        mass += m_trans
        cost += inputs["transition_piece_cost"]
        cg_plat += m_trans * cg_trans

        # Finalize outputs
        cg_plat /= mass
        cb_plat /= volume

        # With CG known, loop back through to compute platform I and added mass
        unit_z = np.array([0.0, 0.0, 1.0])
        I_hull = np.zeros((3, 3))
        m_added = np.zeros(6)
        for k in range(n_member):
            xyz_k = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz_k[:, 0] == NULL)[0][0]
            xyz_k = xyz_k[:inodes, :]

            imass = inputs[f"member{k}:total_mass"]
            cg_k = inputs[f"member{k}:center_of_mass"]
            R = cg_plat - cg_k

            # Figure out angle to make member parallel to global c.s.
            vec_k = xyz_k[-1, :] - xyz_k[0, :]
            T = util.rotate_align_vectors(vec_k, unit_z)

            # Rotate member inertia tensor
            I_k = util.assembleI(inputs[f"member{k}:I_total"])
            I_k_rot = T @ I_k @ T.T

            # Now do parallel axis theorem
            I_hull += np.array(I_k_rot) + imass * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

            # Added mass
            m_add_k = inputs[f"member{k}:added_mass"]
            m_added[:3] += m_add_k[:3]
            m_added[3:] += m_add_k[3:] + np.diag(m_add_k[0] * (np.dot(R, R) * np.eye(3) - np.outer(R, R)))

        # Add in transition piece
        R = cg_plat - cg_trans
        I_hull += I_trans + m_trans * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        # Store outputs
        nelem = elem_A.size
        outputs["platform_elem_D"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_a"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_b"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_t"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_A"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Asx"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Asy"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Ixx"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Iyy"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_J0"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_rho"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_E"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_G"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_TorsC"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_sigma_y"] = NULL * np.ones(NELEM_MAX)

        outputs["platform_elem_D"][:nelem] = elem_D
        outputs["platform_elem_a"][:nelem] = elem_a
        outputs["platform_elem_b"][:nelem] = elem_b
        outputs["platform_elem_t"][:nelem] = elem_t
        outputs["platform_elem_A"][:nelem] = elem_A
        outputs["platform_elem_Asx"][:nelem] = elem_Asx
        outputs["platform_elem_Asy"][:nelem] = elem_Asy
        outputs["platform_elem_Ixx"][:nelem] = elem_Ixx
        outputs["platform_elem_Iyy"][:nelem] = elem_Iyy
        outputs["platform_elem_J0"][:nelem] = elem_J0
        outputs["platform_elem_rho"][:nelem] = elem_rho
        outputs["platform_elem_E"][:nelem] = elem_E
        outputs["platform_elem_G"][:nelem] = elem_G
        outputs["platform_elem_TorsC"][:nelem] = elem_TorsC
        outputs["platform_elem_sigma_y"][:nelem] = elem_sigy
        discrete_outputs["platform_elem_memid"] = elem_memid

        outputs["platform_ballast_mass"] = m_ball
        outputs["platform_hull_mass"] = mass - m_ball
        outputs["platform_cost"] = cost
        outputs["platform_displacement"] = volume
        outputs["platform_hull_center_of_mass"] = cg_plat
        outputs["platform_center_of_buoyancy"] = cb_plat
        outputs["platform_I_hull"] = util.unassembleI(I_hull)
        outputs["platform_Awater"] = Awater
        outputs["platform_Iwaterx"] = Iwaterx
        outputs["platform_Iwatery"] = Iwatery
        outputs["platform_added_mass"] = m_added
        outputs["platform_variable_capacity"] = variable_capacity


class PlatformTurbineSystem(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]
        n_attach = opt["mooring"]["n_attach"]

        self.add_input("platform_hull_center_of_mass", np.zeros(3), units="m")
        self.add_input("platform_ballast_mass", 0.0, units="kg")
        self.add_input("platform_hull_mass", 0.0, units="kg")
        self.add_input("platform_I_hull", np.zeros(6), units="kg*m**2")
        self.add_input("platform_displacement", 0.0, units="m**3")

        self.add_input("turbine_mass", 0.0, units="kg")
        self.add_input("turbine_cg", np.zeros(3), units="m")
        self.add_input("turbine_I", np.zeros(6), units="kg*m**2")
        self.add_input("transition_node", np.zeros(3), units="m")

        self.add_input("rho_water", 0.0, units="kg/m**3")
        self.add_input("mooring_neutral_load", np.zeros((n_attach, 3)), units="N")
        self.add_input("platform_variable_capacity", np.zeros(n_member), units="m**3")

        for k in range(n_member):
            self.add_input(f"member{k}:nodes_xyz", NULL * np.ones((MEMMAX, 3)), units="m")
            self.add_input(f"member{k}:variable_ballast_Vpts", val=np.zeros(10), units="m**3")
            self.add_input(f"member{k}:variable_ballast_spts", val=np.zeros(10))

        self.add_output("system_structural_center_of_mass", np.zeros(3), units="m")
        self.add_output("system_structural_mass", 0.0, units="kg")
        self.add_output("system_center_of_mass", np.zeros(3), units="m")
        self.add_output("system_mass", 0.0, units="kg")
        self.add_output("system_I", np.zeros(6), units="kg*m**2")
        self.add_output("variable_ballast_mass", 0.0, units="kg")
        self.add_output("variable_center_of_mass", val=np.zeros(3), units="m")
        self.add_output("variable_I", np.zeros(6), units="kg*m**2")
        self.add_output("constr_variable_margin", val=0.0)
        self.add_output("member_variable_volume", val=np.zeros(n_member), units="m**3")
        self.add_output("member_variable_height", val=np.zeros(n_member))
        self.add_output("platform_mass", 0.0, units="kg")
        self.add_output("platform_total_center_of_mass", np.zeros(3), units="m")
        self.add_output("platform_I_total", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Mass summaries
        m_platform = inputs["platform_hull_mass"] + inputs["platform_ballast_mass"]
        cg_platform = inputs["platform_hull_center_of_mass"]
        I_platform = util.assembleI(inputs["platform_I_hull"])
        m_turb = inputs["turbine_mass"]
        cg_turb = inputs["turbine_cg"]
        I_turb = util.assembleI(inputs["turbine_I"])
        m_sys = m_platform + m_turb
        outputs["system_structural_mass"] = m_sys

        outputs["system_structural_center_of_mass"] = (m_platform * cg_platform + m_turb * cg_turb) / m_sys

        # Balance out variable ballast
        mooringFz = inputs["mooring_neutral_load"][:, 2].sum()
        capacity = inputs["platform_variable_capacity"]
        capacity_sum = capacity.sum() + EPS  # Avoid divide by zeros
        rho_water = inputs["rho_water"]
        m_variable = inputs["platform_displacement"] * rho_water - m_sys + mooringFz / gravity
        V_variable = m_variable / rho_water
        outputs["variable_ballast_mass"] = m_variable
        outputs["constr_variable_margin"] = V_variable / capacity_sum
        V_variable_member = V_variable * capacity / capacity_sum
        outputs["member_variable_volume"] = V_variable_member
        m_variable_member = V_variable_member * rho_water

        # Now find the CG of the variable mass assigned to each member
        n_member = capacity.size
        outputs["member_variable_height"] = np.zeros(n_member)
        cg_variable_member = np.zeros((n_member, 3))
        for k in range(n_member):
            if V_variable_member[k] == 0.0:
                continue

            xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz[:, 0] == NULL)[0][0]
            xyz = xyz[:inodes, :]
            dxyz = xyz[-1, :] - xyz[0, :]

            spts = inputs[f"member{k}:variable_ballast_spts"]
            Vpts = inputs[f"member{k}:variable_ballast_Vpts"]

            s_cg = np.interp(0.5 * V_variable_member[k], Vpts, spts)
            cg_variable_member[k, :] = xyz[0, :] + s_cg * dxyz

            s_end = np.interp(V_variable_member[k], Vpts, spts)
            outputs["member_variable_height"][k] = s_end - spts[0]

        cg_variable = np.dot(V_variable_member, cg_variable_member) / V_variable
        outputs["variable_center_of_mass"] = cg_variable

        # Now find total system mass
        outputs["platform_mass"] = m_platform + m_variable
        outputs["system_mass"] = m_sys_total = m_sys + m_variable
        outputs["system_center_of_mass"] = cg_sys_total = (
            m_sys * outputs["system_structural_center_of_mass"] + m_variable * cg_variable
        ) / (m_sys + m_variable)

        # Compute the total cg for the platform and the variable ballast together using a weighted sum approach
        cg_plat_total = (m_variable * cg_variable + m_platform * cg_platform) / (m_variable + m_platform)
        outputs["platform_total_center_of_mass"] = cg_plat_total

        # Now loop again to compute variable I
        unit_z = np.array([0.0, 0.0, 1.0])
        I_variable = np.zeros((3, 3))
        for k in range(n_member):
            if V_variable_member[k] == 0.0:
                continue

            xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz[:, 0] == NULL)[0][0]
            xyz = xyz[:inodes, :]
            vec_k = xyz[-1, :] - xyz[0, :]

            ds = outputs["member_variable_height"][k]

            # Compute I aligned with member
            h_k = ds * np.sqrt(np.sum(vec_k**2))
            if h_k == 0.0:
                continue
            r_k = np.sqrt(V_variable_member[k] / h_k / np.pi)
            I_k = (
                m_variable_member[k] * np.r_[(3 * r_k**2 + h_k**2) / 12.0 * np.ones(2), 0.5 * r_k**2, np.ones(3)]
            )

            # Rotate I to global c.s.
            T = util.rotate_align_vectors(vec_k, unit_z)
            I_k_rot = T @ util.assembleI(I_k) @ T.T

            # Now do parallel axis theorem
            R = cg_variable - cg_variable_member[k, :]
            I_variable += np.array(I_k_rot) + m_variable_member[k] * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        # Find platform I with variable contribution
        I_total = np.zeros((3, 3))

        # Compute the full moment of inertia for the platform and variable ballast
        R = cg_plat_total - cg_platform
        I_total += I_platform + m_platform * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        R = cg_plat_total - cg_variable
        I_total += I_variable + m_variable * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["platform_I_total"] = util.unassembleI(I_total)
        outputs["variable_I"] = util.unassembleI(I_variable)

        # Now full system moments of inertia
        I_sys = np.zeros((3, 3))

        R = cg_sys_total - cg_plat_total
        I_sys += I_total + outputs["platform_mass"] * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        R = cg_sys_total - inputs["transition_node"]  # turbine I is already at base, not at its cg
        I_sys += I_turb + m_turb * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["system_I"] = util.unassembleI(I_sys)


class FloatingSystem(om.Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]

        self.add_subsystem("plat", PlatformFrame(options=opt), promotes=["*"])
        self.add_subsystem("mux", PlatformTurbineSystem(options=opt), promotes=["*"])



"""Tests for the `InstallPhase` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import pandas as pd
import pytest
from marmot import Environment

from wisdem.orbit.phases.install import InstallPhase


class BadInstallPhase(InstallPhase):
    """Subclass for testing InstallPhase."""

    def __init__(self, config, **kwargs):
        """Creates an instance of BadInstallPhase."""

        self.config = config
        self.env = Environment()


class SampleInstallPhase(InstallPhase):
    """Subclass for testing InstallPhase."""

    phase = "SampleInstallPhase"

    def __init__(self, config, **kwargs):
        """Creates an instance of SampleInstallPhase."""

        self.config = config
        self.env = Environment()

    def detailed_output(self):
        pass

    def setup_simulation(self):
        pass


base_config = {"port": {"num_cranes": 1, "name": "TEST_PORT"}}


def test_abstract_methods():
    with pytest.raises(TypeError):
        install = BadInstallPhase(base_config)

    install = SampleInstallPhase(base_config)


def test_run():
    sim = SampleInstallPhase(base_config)
    sim.run(until=10)

    assert sim.env.now == 10



"""Tests for the `OffshoreSubstationInstallation` class using feeder barges."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

import pandas as pd
import pytest

from wisdem.orbit import ProjectManager
from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.phases.install import FloatingSubstationInstallation, OffshoreSubstationInstallation
from wisdem.test.test_orbit.data import test_weather
from wisdem.orbit.core.exceptions import MissingComponent

config_single = extract_library_specs("config", "oss_install")
config_floating = extract_library_specs("config", "floating_oss_install")
config_multi = extract_library_specs("config", "oss_install")
config_multi["num_feeders"] = 2


@pytest.mark.parametrize(
    "config",
    (config_single, config_multi),
    ids=["single_feeder", "multi_feeder"],
)
def test_simulation_setup(config):
    sim = OffshoreSubstationInstallation(config)
    assert sim.config == config
    assert sim.env
    assert sim.port
    assert sim.port.crane.capacity == config["port"]["num_cranes"]
    assert sim.num_substations == sim.config["num_substations"]

    assert len(sim.port.items) == 2 * sim.num_substations


def test_floating_simulation_setup():
    sim = FloatingSubstationInstallation(config_floating)
    assert sim.config == config_floating
    assert sim.env
    assert sim.num_substations == sim.config["num_substations"]


@pytest.mark.parametrize(
    "config",
    (config_single, config_multi),
    ids=["single_feeder", "multi_feeder"],
)
def test_vessel_initialization(config):
    sim = OffshoreSubstationInstallation(config)
    assert sim.oss_vessel
    assert sim.oss_vessel.crane

    js = sim.oss_vessel._jacksys_specs
    dp = sim.oss_vessel._dp_specs

    if not any([js, dp]):
        assert False

    for feeder in sim.feeders:
        assert feeder.storage


@pytest.mark.parametrize(
    "config",
    (config_single, config_multi),
    ids=["single_feeder", "multi_feeder"],
)
@pytest.mark.parametrize("weather", (None, test_weather), ids=["no_weather", "test_weather"])
def test_for_complete_logging(weather, config):
    # No weather
    sim = OffshoreSubstationInstallation(config, weather=weather)
    sim.run()

    df = pd.DataFrame(sim.env.actions)
    df = df.assign(shift=(df["time"] - df["time"].shift(1)))

    for vessel in df["agent"].unique():
        _df = df[df["agent"] == vessel].copy()
        _df = _df.assign(shift=(_df["time"] - _df["time"].shift(1)))
        assert (_df["shift"] - _df["duration"]).fillna(0.0).abs().max() < 1e-9

    assert ~df["cost"].isnull().any()
    _ = sim.agent_efficiencies
    _ = sim.detailed_output


@pytest.mark.parametrize("weather", (None, test_weather), ids=["no_weather", "test_weather"])
def test_for_complete_logging_floating(weather):
    sim = FloatingSubstationInstallation(config_floating, weather=weather)
    sim.run()

    df = pd.DataFrame(sim.env.actions)
    df = df.assign(shift=(df["time"] - df["time"].shift(1)))

    for vessel in df["agent"].unique():
        _df = df[df["agent"] == vessel].copy()
        _df = _df.assign(shift=(_df["time"] - _df["time"].shift(1)))
        assert (_df["shift"] - _df["duration"]).fillna(0.0).abs().max() < 1e-9


def test_kwargs():
    sim = OffshoreSubstationInstallation(config_single)
    sim.run()
    baseline = sim.total_phase_time

    keywords = [
        "mono_embed_len",
        "mono_drive_rate",
        "mono_fasten_time",
        "mono_release_time",
        "tp_bolt_time",
        "site_position_time",
        "crane_reequip_time",
        "rov_survey_time",
        "topside_fasten_time",
        "topside_release_time",
    ]

    failed = []

    for kw in keywords:
        default = pt[kw]

        if kw == "mono_drive_rate":
            _new = default - 2

            if _new <= 0:
                raise Exception(f"'{kw}' is less than 0.")

            kwargs = {kw: _new}

        else:
            kwargs = {kw: default + 2}

        new_sim = OffshoreSubstationInstallation(config_single, **kwargs)
        new_sim.run()
        new_time = new_sim.total_phase_time

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"'{failed}' not affecting results.")

    else:
        assert True


def test_kwargs_in_ProjectManager():
    base = deepcopy(config_single)
    base["install_phases"] = ["OffshoreSubstationInstallation"]

    project = ProjectManager(base)
    project.run()
    baseline = project.phase_times["OffshoreSubstationInstallation"]

    keywords = [
        "mono_embed_len",
        "mono_drive_rate",
        "mono_fasten_time",
        "mono_release_time",
        "tp_bolt_time",
        "site_position_time",
        "crane_reequip_time",
        "rov_survey_time",
        "topside_fasten_time",
        "topside_release_time",
    ]

    failed = []

    for kw in keywords:
        default = pt[kw]

        if kw == "mono_drive_rate":
            _new = default - 2

            if _new <= 0:
                raise Exception(f"'{kw}' is less than 0.")

            processes = {kw: _new}

        else:
            processes = {kw: default + 2}

        new_config = deepcopy(base)
        new_config["processes"] = processes

        new_project = ProjectManager(new_config)
        new_project.run()
        new_time = new_project.phase_times["OffshoreSubstationInstallation"]

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"'{failed}' not affecting results.")

    else:
        assert True



import os
import unittest

import openmdao.api as om

# from wisdem.rotorse.rotor_cost import StandaloneRotorCost, initialize_omdao_prob
# from wisdem.glue_code.gc_LoadInputs import WindTurbineOntologyPython
# from wisdem.glue_code.gc_PoseOptimization import PoseOptimization
from wisdem import run_wisdem


class TestRC(unittest.TestCase):
    def testBladeJointSizerBAR_USC(self):
        wisdem_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
        example_dir = os.path.join(wisdem_dir, "examples", "03_blade")  # get path example 03_blade
        fname_modeling_options = os.path.join(example_dir, "modeling_options.yaml")
        fname_wt_input = os.path.join(example_dir, "BAR_USC.yaml")
        fname_analysis_options = os.path.join(example_dir, "analysis_options_no_opt.yaml")
        accuracy = 0
        wt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

        #self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.blade_mass"][0], 50982.52216370957, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.L_transition_joint"][0], -0.004017126893942291, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.t_reinf_ratio_joint"][0], 0.9899571827651443, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.w_reinf_ratio_joint"][0], 1.2087641190997038, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.n_joint_bolt"][0], 38.364778956043914, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.joint_mass"][0], 431.97426980947546, places=accuracy)
        self.assertAlmostEqual(wt_opt["rotorse.rs.bjs.joint_material_cost"][0], 1863.1661079528496, places=accuracy)


if __name__ == "__main__":
    unittest.main()



from __future__ import division

import numpy as np
from scipy.optimize import root, fsolve


def distfunc(dinp, ival=1):
    """
    function for smoothly distributing points along a curve
    according to a set of control points

    Adapted from NN Soerensen's distfunc.f

    parameters:
    -----------
    dinp: list
        list of control points with the form
        [[s0, ds0, n0], [s1, ds1, n1], ... [s<n>, ds<n>, n<n>]]
        where
            s<n> is the curve fraction at each control point,
            ds<n> is the cell size at each control point,
            n<n> is the cell count at each control point.

    returns:
    --------
    s: array
        distribed points
    """

    dinp = np.asarray(dinp)

    ndist = int(dinp[-1, 2])
    nn = len(dinp)
    fdist = np.zeros(ndist)
    dy = np.zeros(ndist)

    fdist[0] = 0.0
    s0 = dinp[0, 0]
    d0 = dinp[0, 1]
    n0 = np.int_(dinp[0, 2]) - 1

    for i in range(1, nn):
        s1 = dinp[i, 0]
        if i > 0 and d0 < 0.0:
            d0 = fdist[n0] - fdist[n0 - 1]
        d1 = dinp[i, 1]
        n1 = np.int_(dinp[i, 2]) - 1
        _len = s1 - s0
        delta1 = d0
        delta2 = d1
        if ival == 1:
            dy[n0 : n1 + 1] = tanhdist(delta1, delta2, _len, n0, n1)
        if ival == 2:
            dy[n0 : n1 + 1] = sinhdist(delta1, delta2, _len, n0, n1)

        for j in range(n0 + 1, n1 + 1):
            fdist[j] = fdist[n0] + dy[j]

        s0 = s1
        d0 = d1
        n0 = n1

    return fdist


def tanhdist(delta1=None, delta2=None, _len=None, i1=None, i2=None):
    if isinstance(delta1, type(np.array([]))):
        delta1 = float(delta1[0])
        delta2 = float(delta1[0])
    ni = i2 - i1
    fdist = np.zeros(ni + 1)
    if i2 == i1:
        return fdist
    delta1 = delta1 / _len
    delta2 = delta2 / _len
    if delta1 <= 0.0 and 1.0 / delta2 < ni:
        delta1 = 1 / (ni**2 * delta2 * 1.02)
    else:
        if delta2 <= 0.0 and 1.0 / delta1 < ni:
            delta2 = 1 / (ni**2 * delta1 * 1.02)
    if delta1 > 0.0 and delta2 > 0.0:
        a = np.sqrt(delta2) / np.sqrt(delta1)
        b = 1.0 / (ni * np.sqrt(delta1 * delta2))
        if b >= 1.0:
            delta = transsinh(b)
            for i in range(ni + 1):
                ftmp = 0.5 * (1 + np.tanh(delta * (i / ni - 0.5)) / np.tanh(0.5 * delta))
                fdist[i] = ftmp / (a + (1 - a) * ftmp)
        else:
            delta = transtanh(b)
            for i in range(ni + 1):
                ftmp = 0.5 * (1 + np.sinh(delta * (i / ni - 0.5)) / np.sinh(0.5 * delta))
                fdist[i] = ftmp / (a + (1 - a) * ftmp)
    else:
        if delta1 > 0.0:
            b = 1.0 / (ni * delta1)
            delta = transsinh(b)
            for i in range(ni + 1):
                fdist[i] = 1.0 + np.tanh(0.5 * delta * (i / ni - 1.0)) / np.tanh(0.5 * delta)
        else:
            if delta2 > 0.0:
                b = 1.0 / (ni * delta2)
                delta = transsinh(b)
                for i in range(ni + 1):
                    fdist[i] = np.tanh(0.5 * delta * i / ni) / np.tanh(0.5 * delta)
            else:
                print("Error from tandist, no cell hight is given")
    for i in range(ni + 1):
        fdist[i] = fdist[i] * _len
    return fdist


def sinhdist(delta1=None, delta2=None, _len=None, i1=None, i2=None):
    if isinstance(delta1, type(np.array([]))):
        delta1 = float(delta1[0])
        delta2 = float(delta1[0])
    ni = i2 - i1
    fdist = np.zeros(ni + 1)
    if i2 == i1:
        return fdist
    delta1 = delta1 / _len
    delta2 = delta2 / _len
    if (delta1 <= 0.0) and ((1 / delta2) < ni):
        delta1 = 1 / (ni**2 * delta2 * 1.02)
    else:
        if (delta2 <= 0.0) and (1 / delta1 < ni):
            delta2 = 1 / (ni**2 * delta1 * 1.02)
    if (delta1 > 0.0) and (delta2 > 0.0):
        a = np.sqrt(delta2) / np.sqrt(delta1)
        b = 1.0 / (ni * np.sqrt(delta1 * delta2))
        if b >= 1.0:
            delta = transsinh(b)
            for i in range(ni + 1):
                ftmp = 0.5 * (1 + np.tanh(delta * (i / ni - 0.5)) / np.tanh(0.5 * delta))
                fdist[i] = ftmp / (a + (1 - a) * ftmp)
        else:
            delta = transtanh(b)
            for i in range(ni + 1):
                ftmp = 0.5 * (1 + np.sinh(delta * (i / ni - 0.5)) / np.sinh(0.5 * delta))
                fdist[i] = ftmp / (a + (1 - a) * ftmp)
    else:
        if delta1 > 0.0:
            b = 1.0 / (ni * delta1)
            delta = transsinh(b)
            for i in range(ni + 1):
                fdist[i] = np.sinh(delta * i / ni) / np.sinh(delta)
        else:
            if delta2 > 0.0:
                b = 1.0 / (ni * delta2)
                delta = transsinh(b)
                for i in range(ni + 1):
                    fdist[i] = 1 - np.sinh(delta * (1 - i / ni)) / np.sinh(delta)
            else:
                disp_([char(" Error from tandist, no cell hight is given ")])
    for i in range(ni):
        fdist[i] = fdist[i] * _len
    return fdist


def transsinh(b):
    """
    solve the transcendental equation

    b=sinh(delta)/delta

    using scipy.optimize.root
    """
    res = root(_transsinh, 1.0, args=(b))

    return res["x"]


def _transsinh(delta, b):
    return delta / (np.sinh(delta) + 1.0 * 10 ** (-60)) - 1.0 / b


def transtanh(b):
    """
    solve the transcendental equation

    b=sinh(delta)/delta

    using scipy.optimize.root
    """
    res = root(_transtanh, 1.0, args=(b))

    return res["x"]


def _transtanh(delta, b):
    return delta / (np.tanh(0.5 * delta) + 1.0 * 10 ** (-60)) - 2.0 / b


def transsinh_(b=None):
    """
    unused fortran translation of Newton-Raphson solver for
    solving the transcendental equation

    b=sinh(delta)/delta
    """
    rlim = 1.0e-6
    nmax = 300
    delta = 1.0
    delta_old = delta
    for n in range(nmax):
        f = delta / (np.sinh(delta) + 1.0 * 10 ** (-60)) - 1.0 / b
        df = (np.sinh(delta) - delta * np.cosh(delta)) / np.sinh(delta) ** 2
        delta = delta - 0.5 * f / df
        if (n == nmax) and (np.abs((delta - delta_old) / delta_old) > rlim):
            disp_([char(" Convergence problem in transsinh dist. function !!! ")])
            disp_([char(" residual "), num2str_((delta - delta_old) / delta_old)])
        delta_old = delta
    return delta


def transtanh_(b=None):
    """
    unused fortran translation of Newton-Raphson solver for
    solving the transcendental equation

    b=tanh(delta)/delta
    """
    if b > 1.0:
        print(" Error form transtanh: Can not be gridded ")
        return delta
    rlim = 1.0e-6
    nmax = 100000
    delta = 1.0
    delta_old = delta
    for n in range(nmax):
        f = delta / (np.tanh(0.5 * delta) + 1.0 * 10 ** (-60)) - 2.0 / b
        df = (np.tanh(0.5 * delta) - delta * (1 - np.tanh(0.5 * delta) ** 2)) / (
            np.tanh(0.5 * delta) ** 2 + 1.0 * 10 ** (-60)
        )
        delta = delta - 0.5 * f / df
        res = np.abs((delta - delta_old) / delta_old)
        if res < rlim:
            return delta
        if (n == nmax) and (res > rlim):
            print(" Convergence problem in transtanh dist. function !!! ")
            print(" residual ", ((delta - delta_old) / delta_old))
            print("n= ", n, "  nmax=", nmax, "   delta=", delta, "   delta_old=", delta_old, "  rlim=", rlim)
        delta_old = delta
    return delta


if __name__ == "__main__":
    d = distfunc([[0, 0.0001, 1], [0.5, 0.001, 30], [1, 0.0001, 100]])
    import matplotlib.pylab as plt

    plt.plot(range(100), d, "-o")
    plt.show()



import traceback
import pandas as pd
import math

from wisdem.landbosse.model.CostModule import CostModule


class SubstationCost(CostModule):
    """
    **SubstationCost.py**

    - Created by Annika Eberle and Owen Roberts on Dec. 17, 2018

    - Refactored by Parangat Bhaskar and Alicia Key on  June 3, 2019

    Calculates the costs associated with substations for land-based wind projects *(module is currently based on curve fit of empirical data)*


    Get project size (project_size_megawatts = num_turbines * turbine_rating_kilowatt / kilowatt_per_megawatt)
    Get interconnect voltage

    Return total substation costs

    \n\n**Keys in the input dictionary are the following:**

    interconnect_voltage_kV
        (int) project interconnection voltage to substation [in kV]

    project_size_megawatts
        (int) total project size [in MW]

    \n\n**Keys in the output dictionary are the following:**

    substation_cost
        (float) cost of substation [in USD]


    """
    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name


    def calculate_costs(self, calculate_costs_input_dict , calculate_costs_output_dict):
        """
        Function to calculate Substation Cost in USD

        Parameters
        -------
        interconnect_voltage_kV
            (in kV)

        project_size_megawatts
            (in MW)


        Returns:
        -------
        substation_cost
            (in USD)

        """

        # Run in utility mode if number of turbines is > 10:
        if calculate_costs_input_dict['num_turbines'] > 10:
            calculate_costs_output_dict['substation_cost_usd'] = 11652 * (
                        calculate_costs_input_dict['interconnect_voltage_kV'] + calculate_costs_input_dict[
                    'project_size_megawatts']) + 11795 * (calculate_costs_input_dict[
                                                              'project_size_megawatts'] ** 0.3549) + 1526800
        # Run in distributed mode if number of turbines is <= 10:
        else:
            calculate_costs_output_dict['substation_cost_usd'] = 0

        calculate_costs_output_dict['substation_cost_output_df'] = pd.DataFrame([['Other', calculate_costs_output_dict['substation_cost_usd'], 'Substation']],
                                                 columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        calculate_costs_output_dict['total_substation_cost'] = calculate_costs_output_dict['substation_cost_output_df']

        return calculate_costs_output_dict['substation_cost_output_df']

    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result = []
        module = type(self).__name__

        for row in self.output_dict['substation_cost_output_df'].itertuples():
            dashed_row = '{} <--> {} <--> {}'.format(row[1], row[3], math.ceil(row[2]))
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'Type of Cost <--> Phase of Construction <--> Cost in USD ',
                'value': dashed_row,
                'last_number': row[2]
            })

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

        self.output_dict['substation_cost_csv'] = result
        return result


    def run_module(self):
        """
        Runs the SubstationCost module and populates the IO dictionaries with calculated values.

         Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.

        """
        try:
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            # self.outputs_for_module_type_operation(self.input_dict, self.output_dict)
            self.output_dict['substation_module_type_operation'] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict['substation_cost_output_df'],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} SubstationCost")
            return 1, error



import copy
import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

import wisdem.towerse.tower as tow
from wisdem.commonse import gravity as g

npts = 100
myones = np.ones((npts,))
RIGID = tow.RIGID


class TestPreDiscretization(unittest.TestCase):
    def testAll(self):
        inputs = {}
        outputs = {}
        inputs["hub_height"] = np.array([ 125.0 ])
        inputs["tower_height"] = np.array([ 100.0 ])
        inputs["foundation_height"] = np.array([ 10.0 ])

        mydis = tow.PreDiscretization()
        mydis.compute(inputs, outputs)
        self.assertEqual(outputs["transition_piece_height"], 10.0)
        self.assertEqual(outputs["z_start"], 10.0)
        npt.assert_array_equal(outputs["joint1"], np.array([0.0, 0.0, 10.0]))
        npt.assert_array_equal(outputs["joint2"], np.array([0.0, 0.0, 110.0]))
        self.assertEqual(outputs["height_constraint"], 15.0)


class TestMass(unittest.TestCase):
    def testAll(self):
        inputs = {}
        outputs = {}
        inputs["joint2"] = np.array([0.0, 0.0, 125.0])
        inputs["rna_mass"] = np.array([ 1000.0 ])
        inputs["rna_I"] = 3e4 * np.ones(6)
        inputs["rna_cg"] = 20.0 * np.ones(3)
        inputs["tower_mass"] = np.array([ 3000.0 ])
        inputs["tower_center_of_mass"] = 0.5 * inputs["joint2"][-1]
        inputs["tower_I_base"] = 2e4 * np.ones(6)

        myobj = tow.TurbineMass()
        myobj.compute(inputs, outputs)

        self.assertEqual(outputs["turbine_mass"], 4e3)
        h = np.r_[0.0, 0.0, 125.0]
        npt.assert_equal(outputs["turbine_center_of_mass"], (1e3 * (inputs["rna_cg"] + h) + 3e3 * 0.5 * h) / 4e3)
        # npt.assert_array_less(5e4, np.abs(outputs["turbine_I_base"]))

        
class TestTowerSE(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        # Store analysis options
        self.modeling_options = {}
        self.modeling_options["materials"] = {}
        self.modeling_options["materials"]["n_mat"] = 1

        self.modeling_options["WISDEM"] = {}
        self.modeling_options["WISDEM"]["n_dlc"] = 1
        self.modeling_options["WISDEM"]["TowerSE"] = {}
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_method"] = "eurocode"
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_length"] = 30.0
        self.modeling_options["WISDEM"]["TowerSE"]["n_layers"] = 1
        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["n_refine"] = 3
        self.modeling_options["WISDEM"]["TowerSE"]["wind"] = "PowerWind"

        self.modeling_options["WISDEM"]["TowerSE"]["gamma_f"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_m"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_n"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_b"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_fatigue"] = 1.0

        # Simplified the options available to the user
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"] = {}
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["shear"] = True
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["geom"] = True
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["modal_method"] = 1
        self.modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["tol"] = 1e-9

    def testProblemLand(self):
        prob = om.Problem(reports=False)
        prob.model = tow.TowerSE(modeling_options=self.modeling_options)
        prob.setup()

        prob["hub_height"] = 80.0

        prob["tower_s"] = np.linspace(0, 1, 3)
        prob["foundation_height"] = 0.0
        prob["tower_height"] = 80.0
        # prob['tower_section_height'] = 40.0*np.ones(2)
        prob["tower_outer_diameter_in"] = 10.0 * np.ones(3)
        prob["tower_layer_thickness"] = 0.1 * np.ones((1, 3))
        prob["outfitting_factor_in"] = 1.0
        prob["tower_layer_materials"] = ["steel"]
        prob["material_names"] = ["steel"]
        prob["E_mat"] = 1e9 * np.ones((1, 3))
        prob["G_mat"] = 1e8 * np.ones((1, 3))
        prob["rho_mat"] = 1e4
        prob["sigma_y_mat"] = 1e8
        prob["sigma_ult_mat"] = 1e8 * np.ones((1, 3))
        prob["wohler_exp_mat"] = 1e1
        prob["wohler_A_mat"] = 1e1

        prob["yaw"] = 0.0
        prob["rna_mass"] = 2e5
        prob["rna_I"] = np.r_[1e5, 1e5, 2e5, np.zeros(3)]
        prob["rna_cg"] = np.array([-3.0, 0.0, 1.0])
        prob["wind_reference_height"] = 80.0
        prob["z0"] = 0.0
        prob["cd_usr"] = -1.0
        prob["rho_air"] = 1.225
        prob["mu_air"] = 1.7934e-5
        prob["shearExp"] = 0.2
        prob["env.Uref"] = 15.0
        prob["tower.rna_F"] = 1e3 * np.arange(2, 5)
        prob["tower.rna_M"] = 1e4 * np.arange(2, 5)
        prob.run_model()

        # All other tests from above
        mass_dens = 1e4 * (5.0**2 - 4.9**2) * np.pi
        npt.assert_equal(prob["z_start"], 0.0)
        npt.assert_equal(prob["transition_piece_height"], 0.0)
        npt.assert_equal(prob["z_param"], np.array([0.0, 40.0, 80.0]))

        self.assertEqual(prob["height_constraint"], 0.0)
        npt.assert_almost_equal(prob["tower_center_of_mass"], 40.0)
        npt.assert_almost_equal(prob["tower_mass"], mass_dens * 80.0)

        npt.assert_equal(prob.model.perf.tower.frame.rnode, np.array([1], dtype=np.int_))
        npt.assert_equal(prob.model.perf.tower.frame.rKx, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.tower.frame.rKy, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.tower.frame.rKz, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.tower.frame.rKtx, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.tower.frame.rKty, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.tower.frame.rKtz, np.array([RIGID]))

        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].NF, np.array([7]))
        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].Fx, np.array([2e3]))
        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].Fy, np.array([3e3]))
        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].Fz, np.array([4e3]))
        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].Mxx, np.array([2e4]))
        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].Myy, np.array([3e4]))
        npt.assert_equal(prob.model.perf.tower.frame.loadCases[0].Mzz, np.array([4e4]))

    def testAddedMassForces(self):
        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = 3

        prob = om.Problem(reports=False)
        prob.model = tow.TowerSE(modeling_options=self.modeling_options)
        prob.setup()

        prob["hub_height"] = 80.0

        prob["tower_s"] = np.linspace(0, 1, 3)
        prob["foundation_height"] = 0.0
        prob["tower_height"] = 60.0
        prob["tower_outer_diameter_in"] = 10.0 * np.ones(3)
        prob["tower_layer_thickness"] = 0.1 * np.ones((1, 3))
        prob["outfitting_factor_in"] = 1.0
        hval = np.array([15.0, 30.0])
        prob["tower_layer_materials"] = ["steel"]
        prob["material_names"] = ["steel"]
        prob["E_mat"] = 1e9 * np.ones((1, 3))
        prob["G_mat"] = 1e8 * np.ones((1, 3))
        prob["rho_mat"] = 1e4
        prob["sigma_y_mat"] = 1e8
        prob["sigma_ult_mat"] = 1e8 * np.ones((1, 3))
        prob["wohler_exp_mat"] = 1e1
        prob["wohler_A_mat"] = 1e1

        prob["yaw"] = 0.0
        # prob["G_soil"] = 1e7
        # prob["nu_soil"] = 0.5
        prob["rna_mass"] = 0.0
        prob["rna_I"] = np.r_[1e5, 1e5, 2e5, np.zeros(3)]
        prob["rna_cg"] = np.array([-3.0, 0.0, 1.0])
        prob["wind_reference_height"] = 80.0
        prob["z0"] = 0.0
        prob["cd_usr"] = -1.0
        prob["rho_air"] = 1.225
        prob["mu_air"] = 1.7934e-5
        prob["shearExp"] = 0.2
        prob["beta_wind"] = 0.0
        prob["env.Uref"] = 15.0
        prob["tower.rna_F"] = 1e3 * np.arange(2, 5)
        prob["tower.rna_M"] = 1e4 * np.arange(2, 5)
        prob.run_model()
        myFz = copy.copy(prob["tower.tower_Fz"])

        prob["tower.rna_F"][-1] += 1e4
        prob.run_model()
        myFz += 1e4
        npt.assert_almost_equal(prob["tower.tower_Fz"], myFz)

    def test15MWmode_shapes(self):
        # --- geometry ----
        h_param = np.array(
            [5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 13.0, 13.0, 13.0, 13.0, 13.0, 13.0, 13.0, 13.0, 13.0, 14.1679]
        )
        d_param = np.array(
            [
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                9.92647687,
                9.44319282,
                8.83283769,
                8.15148167,
                7.38976138,
                6.90908962,
                6.74803581,
                6.57231775,
                6.5,
            ]
        )
        t_param = np.array(
            [
                0.05534138,
                0.05534138,
                0.05344902,
                0.05150928,
                0.04952705,
                0.04751736,
                0.04551709,
                0.0435267,
                0.04224176,
                0.04105759,
                0.0394965,
                0.03645589,
                0.03377851,
                0.03219233,
                0.03070819,
                0.02910109,
                0.02721289,
                0.02400931,
                0.0208264,
                0.02399756,
            ]
        )

        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = len(d_param)

        prob = om.Problem(reports=False)
        prob.model = tow.TowerSE(modeling_options=self.modeling_options)
        prob.setup()

        # Set common and then customized parameters
        prob["hub_height"] = prob["wind_reference_height"] = 30 + 146.1679
        # prob["foundation_height"] = 0.0  # -30.0

        prob["tower_s"] = np.cumsum(np.r_[0.0, h_param]) / h_param.sum()
        prob["foundation_height"] = 0.0  # 15.0
        prob["tower_height"] = h_param.sum()
        prob["tower_outer_diameter_in"] = d_param
        prob["tower_layer_thickness"] = t_param.reshape((1, len(t_param)))
        prob["outfitting_factor_in"] = 1.0
        prob["tower_layer_materials"] = ["steel"]
        prob["material_names"] = ["steel"]
        prob["E_mat"] = 210e9 * np.ones((1, 3))
        prob["G_mat"] = 79.3e9 * np.ones((1, 3))
        prob["rho_mat"] = 7850.0
        prob["sigma_y_mat"] = 345e6
        prob["sigma_ult_mat"] = 500e6 * np.ones((1, 3))
        prob["wohler_exp_mat"] = 4.0
        prob["wohler_A_mat"] = 500e6

        prob["yaw"] = 0.0
        # prob['G_soil'] = 140e6
        # prob['nu_soil'] = 0.4
        prob["shearExp"] = 0.11
        prob["rho_air"] = 1.225
        prob["z0"] = 0.0
        prob["mu_air"] = 1.7934e-5

        mIxx = 379640227.0
        mIyy = 224477294.0
        mIzz = 182971949.0
        mIxy = 0.0
        mIxz = -7259625.38
        mIyz = 0.0
        prob["rna_mass"] = 1007537.0
        prob["rna_I"] = np.array([mIxx, mIyy, mIzz, mIxy, mIxz, mIyz])
        prob["rna_cg"] = np.array([-5.019, 0.0, 0.0])

        prob["env.Uref"] = 0.0  # 20.00138038
        prob["tower.rna_F"] = np.zeros(3)  # np.array([3569257.70891496, -22787.83765441, -404483.54819059])
        prob["tower.rna_M"] = np.zeros(3)  # np.array([68746553.1515807, 16045443.38557568, 1811078.988995])

        # # --- run ---
        prob.run_model()
        """
        Natural Frequencies (Hz): [ 0.2161   0.21842  1.1091   1.167    1.2745   2.3611   2.5877   5.1233  5.2111   9.9725  10.007   10.151   16.388   16.4     18.092   21.813 23.955   23.958   30.184   33.706  ]

        Polynomial fit coefficients to modal displacements (x^2, x^3, x^4, x^5, x^6)
        1st Fore-aft    = [1.11422342, -2.73438505, 6.84397071, -5.97959674, 1.75578766]
        2nd Fore-aft    = [-48.86125831, 82.74454067, -156.79260263, 208.53125496, -84.62193469]
        1st Side-side   = [1.10492357, -2.71587869, 6.80247339, -5.93612744, 1.74460918]
        2nd Side-side   = [48.9719383, -89.25323746, 183.04839183, -226.34534799, 84.57825533]
        """

    def testExampleRegression(self):
        # --- geometry ----
        h_param = np.diff(np.array([0.0, 43.8, 87.6]))
        d_param = np.array([6.0, 4.935, 3.87])
        t_param = 1.3 * np.array([0.027, 0.023, 0.019])
        z_foundation = 0.0
        yaw = 0.0
        Koutfitting = 1.07

        # --- material props ---
        E = 210e9
        G = 80.8e9
        rho = 8500.0
        sigma_y = 450.0e6
        sigma_ult = wohler_A = 500e6
        wohler_exp = 4.0

        # --- extra mass ----
        m = np.array([285598.8])
        mIxx = 1.14930678e08
        mIyy = 2.20354030e07
        mIzz = 1.87597425e07
        mIxy = 0.0
        mIxz = 5.03710467e05
        mIyz = 0.0
        mI = np.array([mIxx, mIyy, mIzz, mIxy, mIxz, mIyz])
        mrho = np.array([-1.13197635, 0.0, 0.50875268])
        # -----------

        # --- wind ---
        wind_zref = 90.0
        wind_z0 = 0.0
        shearExp = 0.2
        cd_usr = -1.0
        # ---------------

        # --- costs ---
        material_cost = 5.0
        labor_cost = 100.0 / 60.0
        painting_cost = 30.0
        # ---------------

        # two load cases.  TODO: use a case iterator

        # # --- loading case 1: max Thrust ---
        wind_Uref1 = 11.73732
        Fx1 = 1284744.19620519
        Fy1 = 0.0
        Fz1 = -2914124.84400512 + m * g
        Mxx1 = 3963732.76208099
        Myy1 = -2275104.79420872
        Mzz1 = -346781.68192839
        # # ---------------

        # # --- loading case 2: max wind speed ---
        wind_Uref2 = 70.0
        Fx2 = 930198.60063279
        Fy2 = 0.0
        Fz2 = -2883106.12368949 + m * g
        Mxx2 = -1683669.22411597
        Myy2 = -2522475.34625363
        Mzz2 = 147301.97023764
        # # ---------------

        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = len(d_param)
        self.modeling_options["WISDEM"]["TowerSE"]["n_layers"] = 1
        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = len(d_param)
        self.modeling_options["WISDEM"]["n_dlc"] = 2
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_f"] = 1.35
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_m"] = 1.3
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_n"] = 1.0
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_b"] = 1.1
        self.modeling_options["WISDEM"]["TowerSE"]["gamma_fatigue"] = 1.35 * 1.3 * 1.0

        def fill_prob():
            prob = om.Problem(reports=False)
            prob.model = tow.TowerSE(modeling_options=self.modeling_options)
            prob.setup()

            if self.modeling_options["WISDEM"]["TowerSE"]["wind"] == "PowerWind":
                prob["shearExp"] = shearExp

            # assign values to params

            # --- geometry ----
            prob["hub_height"] = h_param.sum()
            prob["tower_s"] = np.cumsum(np.r_[0.0, h_param]) / h_param.sum()
            prob["foundation_height"] = z_foundation
            prob["tower_height"] = h_param.sum()
            prob["tower_outer_diameter_in"] = d_param
            prob["tower_layer_thickness"] = t_param.reshape((1, len(t_param)))
            prob["outfitting_factor_in"] = Koutfitting
            prob["tower_layer_materials"] = ["steel"]
            prob["material_names"] = ["steel"]
            prob["yaw"] = yaw
            # --- material props ---
            prob["E_mat"] = E * np.ones((1, 3))
            prob["G_mat"] = G * np.ones((1, 3))
            prob["rho_mat"] = rho
            prob["sigma_y_mat"] = sigma_y
            prob["sigma_ult_mat"] = sigma_ult * np.ones((1, 3))
            prob["wohler_exp_mat"] = wohler_exp
            prob["wohler_A_mat"] = wohler_A

            # --- extra mass ----
            prob["rna_mass"] = m
            prob["rna_I"] = mI
            prob["rna_cg"] = mrho
            # -----------

            # --- costs ---
            prob["unit_cost_mat"] = [material_cost]
            prob["labor_cost_rate"] = labor_cost
            prob["painting_cost_rate"] = painting_cost
            # -----------

            # --- wind & wave ---
            prob["wind_reference_height"] = wind_zref
            prob["z0"] = wind_z0
            prob["cd_usr"] = cd_usr
            prob["rho_air"] = 1.225
            prob["mu_air"] = 1.7934e-5

            # # --- loading case 1: max Thrust ---
            prob["env1.Uref"] = wind_Uref1
            prob["env2.Uref"] = wind_Uref2

            prob["tower.rna_F"] = np.c_[np.r_[Fx1, Fy1, Fz1], np.r_[Fx2, Fy2, Fz2]]
            prob["tower.rna_M"] = np.c_[np.r_[Mxx1, Myy1, Mzz1], np.r_[Mxx2, Myy2, Mzz2]]
            # # ---------------

            return prob

        # # --- run ---
        prob = fill_prob()
        prob.run_model()

        npt.assert_almost_equal(prob["z_full"], [0.0, 14.6, 29.2, 43.8, 58.4, 73.0, 87.6])
        npt.assert_almost_equal(prob["outer_diameter_full"], [6.0, 5.645, 5.29, 4.935, 4.58, 4.225, 3.87])
        npt.assert_almost_equal(prob["t_full"], [0.0325, 0.0325, 0.0325, 0.0273, 0.0273, 0.0273])

        npt.assert_almost_equal(prob["tower_mass"], [370541.14008246])
        npt.assert_almost_equal(prob["tower_center_of_mass"], [38.78441074])
        npt.assert_almost_equal(prob["constr_d_to_t"], [168.23076923, 161.26373626])
        npt.assert_almost_equal(prob["constr_taper"], [0.8225, 0.78419453])
        npt.assert_almost_equal(prob["env1.Uref"], [11.73732])
        npt.assert_almost_equal(prob["env2.Uref"], [70.0])
        npt.assert_almost_equal(prob["tower.f1"], [0.34560], 5)
        npt.assert_almost_equal(prob["tower.top_deflection"], [0.66837, 0.4749321])
        npt.assert_almost_equal(
            prob["post.constr_stress"].T,
            [
                [0.37859014, 0.34400647, 0.29456394, 0.2663071 , 0.15248295, 0.06494359],
                [0.26212952, 0.23803754, 0.2030959 , 0.18133974, 0.09896319, 0.04456172],
            ],
        )
        npt.assert_almost_equal(
            prob["post.constr_global_buckling"].T,
            [
                [0.4886788, 0.4503055, 0.3971611, 0.3689947, 0.252642, 0.1473065],
                [0.3953861, 0.3657697, 0.3243708, 0.3011636, 0.2097029, 0.1316665],
            ],
        )
        npt.assert_almost_equal(
            prob["post.constr_shell_buckling"].T,
            [
                [0.23067874, 0.18651124, 0.13545567, 0.12620337, 0.04734878, 0.01360982],
                [0.11961036, 0.09709062, 0.07034931, 0.06611991, 0.02478743, 0.00959923],
            ],
        )
        npt.assert_almost_equal(
            prob["tower.turbine_F"].T, [[1.28474420e6, 0.0, -3.76200522e6], [9.30198601e5, 0.0, -4.39569591e6]], 2
        )
        npt.assert_almost_equal(
            prob["tower.turbine_M"].T,
            [[4.01664296e06, 1.11242143e08, -3.46781682e05], [-1.71083379e06, 7.98010157e07, 1.47301970e05]],
            -2,
        )

        # Now regression on DNV-GL C202 methods
        self.modeling_options["WISDEM"]["TowerSE"]["buckling_method"] = "dnvgl"
        prob = fill_prob()
        prob.run_model()

        npt.assert_almost_equal(prob["z_full"], [0.0, 14.6, 29.2, 43.8, 58.4, 73.0, 87.6])
        npt.assert_almost_equal(prob["outer_diameter_full"], [6.0, 5.645, 5.29, 4.935, 4.58, 4.225, 3.87])
        npt.assert_almost_equal(prob["t_full"], [0.0325, 0.0325, 0.0325, 0.0273, 0.0273, 0.0273])

        npt.assert_almost_equal(prob["tower_mass"], [370541.14008246])
        npt.assert_almost_equal(prob["tower_center_of_mass"], [38.78441074])
        npt.assert_almost_equal(prob["constr_d_to_t"], [168.23076923, 161.26373626])
        npt.assert_almost_equal(prob["constr_taper"], [0.8225, 0.78419453])
        npt.assert_almost_equal(prob["env1.Uref"], [11.73732])
        npt.assert_almost_equal(prob["env2.Uref"], [70.0])
        npt.assert_almost_equal(prob["tower.f1"], [0.34560], 5)
        npt.assert_almost_equal(prob["tower.top_deflection"], [0.66837, 0.4749321])
        npt.assert_almost_equal(
            prob["post.constr_stress"].T,
            [
                [0.37859014, 0.34400647, 0.29456394, 0.2663071 , 0.15248295, 0.06494359],
                [0.26212952, 0.23803754, 0.2030959 , 0.18133974, 0.09896319, 0.04456172],
            ],
        )
        npt.assert_almost_equal(
            prob["post.constr_global_buckling"].T,
            [
                [0.61035748, 0.55938751, 0.47287421, 0.4249917 , 0.23820016, 0.074166  ],
                [0.45451296, 0.41435613, 0.3476131 , 0.30829985, 0.16478585, 0.04921057],
            ],
        )
        npt.assert_almost_equal(
            prob["post.constr_shell_buckling"].T,
            [
                [0.05559682, 0.05320304, 0.05133924, 0.07092966, 0.07038021, 0.070811  ],
                [0.05057771, 0.05009383, 0.04746213, 0.0651608 , 0.06238187, 0.06025312],
            ],
        )
        npt.assert_almost_equal(
            prob["tower.turbine_F"].T, [[1.28474420e6, 0.0, -3.76200522e6], [9.30198601e5, 0.0, -4.39569591e6]], 2
        )
        npt.assert_almost_equal(
            prob["tower.turbine_M"].T,
            [[4.01664296e06, 1.11242143e08, -3.46781682e05], [-1.71083379e06, 7.98010157e07, 1.47301970e05]],
            -2,
        )


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestPreDiscretization),
        unittest.TestLoader().loadTestsFromTestCase(TestMass),
        unittest.TestLoader().loadTestsFromTestCase(TestTowerSE),
    ]
    return unittest.TestSuite(suite)

if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



"""Tests for the `ExportSystemDesign` class."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"

from copy import deepcopy

import pytest

from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.phases.design import ExportSystemDesign

config = extract_library_specs("config", "export_design")


def test_export_system_creation():
    export = ExportSystemDesign(config)
    export.run()

    assert export.num_cables
    assert export.length
    assert export.mass
    assert export.cable
    assert export.total_length
    assert export.total_mass


def test_number_cables():
    export = ExportSystemDesign(config)
    export.run()

    assert export.num_cables == 11


def test_cable_length():
    export = ExportSystemDesign(config)
    export.run()

    length = (0.02 + 3 + 30) * 1.01
    assert export.length == length


def test_cable_mass():
    export = ExportSystemDesign(config)
    export.run()

    length = (0.02 + 3 + 30) * 1.01
    mass = length * export.cable.linear_density
    assert export.mass == mass


def test_total_cable():
    export = ExportSystemDesign(config)
    export.run()

    length = 0.02 + 3 + 30
    length += length * 0.01
    mass = length * export.cable.linear_density
    assert export.total_mass == pytest.approx(mass * 11, abs=1e-10)
    assert export.total_length == pytest.approx(length * 11, abs=1e-10)


def test_cables_property():
    export = ExportSystemDesign(config)
    export.run()

    assert (export.sections_cables == export.cable.name).sum() == export.num_cables


def test_cable_lengths_property():
    export = ExportSystemDesign(config)
    export.run()

    cable_name = export.cable.name
    assert (export.cable_lengths_by_type[cable_name] == export.length).sum() == export.num_cables


def test_total_cable_len_property():
    export = ExportSystemDesign(config)
    export.run()

    cable_name = export.cable.name
    assert export.total_cable_length_by_type[cable_name] == pytest.approx(export.total_length, abs=1e-10)


def test_design_result():
    export = ExportSystemDesign(config)
    export.run()

    _ = export.cable.name
    cables = export.design_result["export_system"]["cable"]

    assert cables["sections"] == [export.length]
    assert cables["number"] == 11
    assert cables["linear_density"] == export.cable.linear_density


def test_floating_length_calculations():
    base = deepcopy(config)
    base["site"]["depth"] = 250
    base["export_system_design"]["touchdown_distance"] = 0

    sim = ExportSystemDesign(base)
    sim.run()

    base_length = sim.total_length

    with_cat = deepcopy(config)
    with_cat["site"]["depth"] = 250

    new = ExportSystemDesign(with_cat)
    new.run()

    assert new.total_length < base_length



import numpy as np
from scipy.optimize import fsolve


class CylinderBuckling:
    """
    Class for calculating cylinder buckling strengths based on DNVGL-RP-C2020.

    Supports the following cylinder configurations:
    - Ring stiffened
    - Longitudinal stiffened
    - Orthogonally stiffened

    Currently supports calculations for:
    - Shell buckling
    """

    def __init__(
        self,
        l,
        d,
        t,
        ring_stiffeners=False,
        num_longitudinal=0,
        E=200e9,
        G=79.3e9,
        sigma_y=345e6,
        gamma_m=1.15,
        mod_length=0.0,
        **kwargs,
    ):
        """
        Creates an instance of `CylinderBuckling`.

        Parameters
        ----------
        z : np.array | list
            Section positions (m).
        d : np.array | list
            Diameters at cylinder points defined in L (m2).
        t : np.array | list
            Thicknesses at cylinder points defined in L (m2).
        ring_stiffeners : bool
            Toggle ring stiffeners at section changes.
            Default: False
        num_longitudinal : int
            Number of longitudinal stiffeners.
            Default: 0
        E : int | float
            Isotropic Young's modulus (Pa).
            Default: 200e9
        G : int | float
            Shear modulus (Pa).
            Default: 79.39e9
        sigma_y : int | float
            Yield Stress (Pa).
            Default: 345e6
        gamma_m : int | float
            Material factor.
        """

        self._l = np.array(l)
        self._d = np.array(d)
        self._t = np.array(t)
        self.E = E
        self.G = G
        self.sigma_y = sigma_y
        self.gamma_m = gamma_m
        self.mod_length = mod_length

        self.A = kwargs.get("A", np.zeros(len(self.t)))
        self.I = kwargs.get("I", np.zeros(len(self.t)))

        self._ring = ring_stiffeners
        self._long = num_longitudinal

    @property
    def d(self):
        """Section diameters at midpoints (m)."""
        if self._d.size == self._l.size + 1:
            return 0.5 * (self._d[1:] + self._d[:-1])
        elif self._d.size == self._l.size:
            return self._d
        else:
            raise ValueError(f"Incompatible size for diameter. Expected: {self._t.size + 1} vs {self._d.size}")

    @property
    def r(self):
        """Section radii at midpoints (m)."""

        return 0.5 * self.d

    @property
    def t(self):
        """Section thicknesses at midpoints (m)."""
        if self._l.size == self._t.size:
            return self._t
        elif self._l.size + 1 == self._t.size:
            return 0.5 * (self._t[1:] + self._t[:-1])
        else:
            raise ValueError(f"Incompatible size for thickness. Expected: {self._d.size - 1} vs {self._t.size}")

    @property
    def te(self):
        """Equivalent section thickness (m). Only applicable for longitudinal membrane stress."""

        if self.s is False:
            return self.t

        return self.t + self.A / self.s

    @property
    def Ac(self):
        """Cross sectional area of complete cylinder section"""
        Ac = (self.d**2 - (self.d - 2 * self.te) ** 2) * np.pi / 4.0
        Ac[self.A>0] = self.A[self.A>0]
        return Ac

    @property
    def Ic(self):
        """Cross sectional area of complete cylinder section"""
        Ic = (self.d**4 - (self.d - 2 * self.te) ** 4) * np.pi / 64.0
        Ic[self.I>0] = self.I[self.I>0]
        return Ic

    @property
    def s(self):
        """Section distance between longitudinal stiffeners (m)."""

        if self._long == 0:
            return False

        return np.pi * self.d / self._long

    @property
    def l(self):
        """Section heights (m)."""
        return self._l

    @property
    def v(self):
        """Poisson ratio for isotropic materials."""

        return 0.5 * self.E / self.G - 1.0

    @property
    def curvature_parameter(self):
        """"""

        if self._long:
            return ((self.s**2) / (self.r * self.t)) * np.sqrt(1 - self.v**2)

        else:
            return ((self.l**2) / (self.r * self.t)) * np.sqrt(1 - self.v**2)

    @property
    def Z(self):
        """Shorthand for `self.curvature_parameter`."""

        return self.curvature_parameter

    ## Buckling Coefficients
    # Ring Stiffened (unstiffened cylinder)
    def _C_axial_bending_unstiffened(self):
        """See Section 3.4, Table 3-2."""

        psi = 1
        xi = 0.702 * self.Z
        rho = 0.5 * (1 + self.r / (150 * self.t)) ** -0.5
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    def _C_torsion_unstiffened(self):
        """See Section 3.4, Table 3-2."""

        psi = 5.34
        xi = 0.856 * self.Z**0.75
        rho = 0.6
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    def _C_lateral_unstiffened(self):
        """See Section 3.4, Table 3-2."""

        psi = 4
        xi = 1.04 * np.sqrt(self.Z)
        rho = 0.6
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    def _C_hydrostatic_unstiffened(self):
        """See Section 3.4, Table 3-2."""

        psi = 2
        xi = 1.04 * np.sqrt(self.Z)
        rho = 0.6
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    # Longitudinal Stiffened and Orthongonally Stiffened
    def _C_axial_longitudinal(self):
        """See Section 3.3, Table 3-1."""

        if self.s is False:
            raise ValueError("Cylinder is not longitudinally stiffened. Incorrect use of this equation.")

        psi = 4
        xi = 0.702 * self.Z
        rho = 0.5 * (1 + self.r / (150.0 * self.t)) ** -0.5
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    def _C_shear_longitudinal(self):
        """See Section 3.3, Table 3-1."""

        if self.s is False:
            raise ValueError("Cylinder is not longitudinally stiffened. Incorrect use of this equation.")

        psi = 5.34 + 4 * (self.s / self.l) ** 2
        xi = 0.856 * np.sqrt(self.s / self.l) * self.Z**0.75
        rho = 0.6
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    def _C_compression_longitudinal(self):
        """See Section 3.3, Table 3-1."""

        if self.s is False:
            raise ValueError("Cylinder is not longitudinally stiffened. Incorrect use of this equation.")

        psi = (1 + (self.s / self.l) ** 2) ** 2
        xi = 1.04 * (self.s / self.l) * np.sqrt(self.Z)
        rho = 0.6
        return psi * np.sqrt(1 + (rho * xi / psi) ** 2)

    # Strength Calculations
    @property
    def fea(self):
        """Characteristic buckling strength due to axial or bending loads."""

        if self._long:
            return self._shell_buckling_strength(self._C_axial_longitudinal())

        else:
            return self._shell_buckling_strength(self._C_axial_bending_unstiffened())

    @property
    def fet(self):
        """Characteristic buckling strength due to torsion."""

        if self._long:
            return self._shell_buckling_strength(self._C_shear_longitudinal())

        else:
            return self._shell_buckling_strength(self._C_torsion_unstiffened())

    @property
    def feh(self):
        """Charcteristic buckling strength due to circumferntial compression."""

        if self._long:
            return self._shell_buckling_strength(self._C_compression_longitudinal())

        else:
            return self._shell_buckling_strength(self._C_lateral_unstiffened())

    def _shell_buckling_strength(self, C):
        """"""

        if self._long:
            return ((C * np.pi**2 * self.E) / (12 * (1 - self.v**2))) * (self.t / self.s) ** 2

        else:
            return ((C * np.pi**2 * self.E) / (12 * (1 - self.v**2))) * (self.t / self.l) ** 2

    def von_mises(self, axial, hoop, shear):
        a = ((axial + hoop) / 2.0) ** 2
        b = ((axial - hoop) / 2.0) ** 2
        c = shear**2

        return np.sqrt(a + 3.0 * (b + c))

    # Utilizations
    def run_buckling_checks(self, Fz, M, sigma_x, sigma_h, sigma_t):
        """

        Parameters
        ----------
        sigma_a :
        sigma_h :
        sigma_t :
        """

        # Axial stress contributions
        sigma_a = Fz / (2 * np.pi * self.r * self.te)
        sigma_m = M * self.r / self.Ic

        shell_utilization, material_factors = self.utilization_shell(sigma_a, sigma_m, sigma_h, sigma_t)
        fak = self.get_fak(sigma_a, sigma_m, sigma_h, sigma_t)
        global_utilization = self.utilization_global(sigma_a, sigma_m, fak)

        return {"Shell": shell_utilization, "Global": global_utilization}

    def utilization_shell(self, _axial, _bending, _hoop, _shear):
        # Set Positive Stresses Equal to 0. See 3.2.4 - 3.2.6
        axial = np.abs(np.minimum(_axial, 0.0))
        bending = np.abs(np.minimum(_bending, 0.0))
        hoop = np.abs(np.minimum(_hoop, 0.0))
        shear = np.abs(_shear)

        eps = 1e-7
        vm = self.von_mises(axial, hoop, shear) + eps

        lambda_s = np.sqrt((self.sigma_y / vm) * ((axial + bending) / self.fea + shear / self.fet + hoop / self.feh))

        gamma_m = 0.85 + 0.6 * lambda_s
        gamma_m[lambda_s < 0.5] = 1.15
        gamma_m[lambda_s >= 1.0] = 1.45

        fks = self.sigma_y / np.sqrt(1 + lambda_s**4.0)
        fksd = fks / np.array(gamma_m)

        shell_util = vm / fksd

        return shell_util, gamma_m

    def get_fak(self, axial, bending, hoop, shear):
        fak = fsolve(self._fak_wrapper, -np.abs(axial), args=(bending, hoop, shear), xtol=1e-4, maxfev=50)

        # Using mean instead of min here for more stable performance in optimization
        return np.abs(fak).mean()

    def _fak_wrapper(self, axial, *data):
        bending, hoop, shear = data
        util, _ = self.utilization_shell(axial, bending, hoop, shear)

        return util - 1

    def utilization_global(self, _sigma_a, sigma_m, fak):
        """"""
        # This is column buckling strength in DNVGL
        # Stresses
        sigma_a = np.abs(np.minimum(_sigma_a, 0.0))
        # sigma_h = np.abs(np.minimum(_sigma_h, 0.0))

        # Euler buckling strength
        k = 2.0  # Fixed-free
        L = self._l.sum() - self.mod_length

        fE = (np.pi**2 * self.E * self.Ic) / ((k * L) ** 2 * self.Ac)

        # ls
        lambda_s = k * L / (np.pi * np.sqrt(self.Ic / self.Ac)) * np.sqrt(fak / self.E)

        gamma_m = 0.85 + 0.6 * lambda_s
        gamma_m[lambda_s < 0.5] = 1.15
        gamma_m[lambda_s >= 1.0] = 1.45

        # fak
        fakd = fak / gamma_m

        fkc = (1 - 0.28 * lambda_s**2)
        fkc[lambda_s > 1.34] = 0.9 / (lambda_s[lambda_s > 1.34] ** 2)
        fkc = fkc * fak
        fkcd = np.array(fkc / gamma_m)

        util = sigma_a / fkcd + 1 / fakd * (sigma_m / (1 - sigma_a / fE))
        return util



#!/usr/bin/env python
# encoding: utf-8
"""
csystem.py

Created by Andrew Ning on 2/21/2012.
Copyright (c) NREL. All rights reserved.
"""

from __future__ import print_function

import numpy as np


class DirectionVector(object):
    """Handles rotation of direction vectors to appropriate coordinate systems.
    All angles must be in degrees.

    """

    def __init__(self, x, y, z, dx=None, dy=None, dz=None):
        """3-Dimensional vector that depends on direction only (not position).

        Parameters
        ----------
        x : float or ndarray
            x-direction of vector(s)
        y : float or ndarray
            y-direction of vector(s)
        z : float or ndarray
            z-direction of vector(s)

        """

        self.x = np.array(x)
        self.y = np.array(y)
        self.z = np.array(z)

        if dx is None:
            dx = {}
            dx["dx"] = np.ones_like(self.x)
            dx["dy"] = np.zeros_like(self.y)
            dx["dz"] = np.zeros_like(self.z)

            dy = {}
            dy["dx"] = np.zeros_like(self.x)
            dy["dy"] = np.ones_like(self.y)
            dy["dz"] = np.zeros_like(self.z)

            dz = {}
            dz["dx"] = np.zeros_like(self.x)
            dz["dy"] = np.zeros_like(self.y)
            dz["dz"] = np.ones_like(self.z)

        self.dx = dx
        self.dy = dy
        self.dz = dz

    @classmethod
    def fromArray(cls, array):
        """initialize with NumPy array

        Parameters
        ----------
        array : ndarray
            construct DirectionVector using array of size 3

        """

        return cls(array[0], array[1], array[2])

    def toArray(self):
        """convert DirectionVector to NumPy array

        Returns
        -------
        array : ndarray
            NumPy array in order x, y, z containing DirectionVector data

        """

        return np.c_[self.x, self.y, self.z]

    def _rotateAboutZ(self, xstring, ystring, zstring, theta, thetaname, reverse=False):
        """
        x X y = z.  rotate c.s. about z, +theta
        all angles in degrees
        """

        thetaM = 1.0
        if reverse:
            thetaM = -1.0

        x = getattr(self, xstring)
        y = getattr(self, ystring)
        z = getattr(self, zstring)
        dx = getattr(self, "d" + xstring)
        dy = getattr(self, "d" + ystring)
        dz = getattr(self, "d" + zstring)

        theta = np.radians(theta * thetaM)
        c = np.cos(theta)
        s = np.sin(theta)

        xnew = x * c + y * s
        ynew = -x * s + y * c
        znew = z

        angles = []
        for key in dx.keys():
            if not key in ["dx", "dy", "dz"]:
                angles.append(key)

        dxnew = {}
        dxnew["dx"] = dx["dx"] * c + dy["dx"] * s
        dxnew["dy"] = dx["dy"] * c + dy["dy"] * s
        dxnew["dz"] = dx["dz"] * c + dy["dz"] * s
        dxnew["d" + thetaname] = (-x * s + y * c) * np.radians(thetaM)
        for dangle in angles:
            dxnew[dangle] = dx[dangle] * c + dy[dangle] * s

        dynew = {}
        dynew["dx"] = -dx["dx"] * s + dy["dx"] * c
        dynew["dy"] = -dx["dy"] * s + dy["dy"] * c
        dynew["dz"] = -dx["dz"] * s + dy["dz"] * c
        dynew["d" + thetaname] = (-x * c - y * s) * np.radians(thetaM)
        for dangle in angles:
            dynew[dangle] = -dx[dangle] * s + dy[dangle] * c

        dznew = {}
        dznew["dx"] = dz["dx"] * np.ones_like(theta)  # multiply by ones just to get right size in case of float
        dznew["dy"] = dz["dy"] * np.ones_like(theta)
        dznew["dz"] = dz["dz"] * np.ones_like(theta)
        dznew["d" + thetaname] = np.zeros_like(theta)
        for dangle in angles:
            dznew[dangle] = dz[dangle]

        return xnew, ynew, znew, dxnew, dynew, dznew

    def windToInertial(self, beta):
        """Rotates from wind-aligned to inertial

        Parameters
        ----------
        beta : float (deg)
            wind angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the inertial coordinate system

        """
        xw, yw, zw, dxw, dyw, dzw = self._rotateAboutZ("x", "y", "z", beta, "beta", reverse=True)
        return DirectionVector(xw, yw, zw, dxw, dyw, dzw)

    def inertialToWind(self, beta):
        """Rotates from inertial to wind-aligned

        Parameters
        ----------
        beta : float (deg)
            wind angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the wind-aligned coordinate system

        """
        xw, yw, zw, dxw, dyw, dzw = self._rotateAboutZ("x", "y", "z", beta, "beta")
        return DirectionVector(xw, yw, zw, dxw, dyw, dzw)

    def yawToWind(self, Psi):
        """Rotates from yaw-aligned to wind-aligned

        Parameters
        ----------
        Psi : float (deg)
            yaw angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the wind-aligned coordinate system

        """
        xw, yw, zw, dxw, dyw, dzw = self._rotateAboutZ("x", "y", "z", Psi, "yaw", reverse=True)
        return DirectionVector(xw, yw, zw, dxw, dyw, dzw)

    def windToYaw(self, Psi):
        """Rotates from wind-aligned to yaw-aligned

        Parameters
        ----------
        Psi : float (deg)
            yaw angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the yaw-aligned coordinate system

        """
        xy, yy, zy, dxy, dyy, dzy = self._rotateAboutZ("x", "y", "z", Psi, "yaw")
        return DirectionVector(xy, yy, zy, dxy, dyy, dzy)

    def hubToYaw(self, Theta, derivatives=False):
        """Rotates from hub-aligned to yaw-aligned

        Parameters
        ----------
        Theta : float (deg)
            tilt angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the yaw-aligned coordinate system

        """
        zy, xy, yy, dzy, dxy, dyy = self._rotateAboutZ("z", "x", "y", Theta, "tilt", reverse=True)
        return DirectionVector(xy, yy, zy, dxy, dyy, dzy)

    def yawToHub(self, Theta):
        """Rotates from yaw-aligned to hub-aligned

        Parameters
        ----------
        Theta : float (deg)
            tilt angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the hub-aligned coordinate system

        """
        zh, xh, yh, dzh, dxh, dyh = self._rotateAboutZ("z", "x", "y", Theta, "tilt")
        return DirectionVector(xh, yh, zh, dxh, dyh, dzh)

    def hubToAzimuth(self, Lambda):
        """Rotates from hub-aligned to azimuth-aligned

        Parameters
        ----------
        Lambda : float or ndarray (deg)
            azimuth angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the azimuth-aligned coordinate system

        """
        yz, zz, xz, dyz, dzz, dxz = self._rotateAboutZ("y", "z", "x", Lambda, "azimuth")
        return DirectionVector(xz, yz, zz, dxz, dyz, dzz)

    def azimuthToHub(self, Lambda):
        """Rotates from azimuth-aligned to hub-aligned

        Parameters
        ----------
        Lambda : float or ndarray (deg)
            azimuth angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the hub-aligned coordinate system

        """

        yh, zh, xh, dyh, dzh, dxh = self._rotateAboutZ("y", "z", "x", Lambda, "azimuth", reverse=True)
        return DirectionVector(xh, yh, zh, dxh, dyh, dzh)

    def azimuthToBlade(self, Phi):
        """Rotates from azimuth-aligned to blade-aligned

        Parameters
        ----------
        Phi : float (deg)
            precone angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the blade-aligned coordinate system

        """

        zb, xb, yb, dzb, dxb, dyb = self._rotateAboutZ("z", "x", "y", Phi, "precone", reverse=True)
        return DirectionVector(xb, yb, zb, dxb, dyb, dzb)

    def bladeToAzimuth(self, Phi):
        """Rotates from blade-aligned to azimuth-aligned

        Parameters
        ----------
        Phi : float (deg)
            precone angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the azimuth-aligned coordinate system


        """

        za, xa, ya, dza, dxa, dya = self._rotateAboutZ("z", "x", "y", Phi, "precone")
        return DirectionVector(xa, ya, za, dxa, dya, dza)

    def airfoilToBlade(self, theta):
        """Rotates from airfoil-aligned to blade-aligned

        Parameters
        ----------
        theta : float (deg)
            twist angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the blade-aligned coordinate system

        """

        xb, yb, zb, dxb, dyb, dzb = self._rotateAboutZ("x", "y", "z", theta, "theta")
        return DirectionVector(xb, yb, zb, dxb, dyb, dzb)

    def bladeToAirfoil(self, theta):
        """Rotates from blade-aligned to airfoil-aligned

        Parameters
        ----------
        theta : float (deg)
            twist angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the airfoil-aligned coordinate system

        """

        xa, ya, za, dxa, dya, dza = self._rotateAboutZ("x", "y", "z", theta, "theta", reverse=True)
        return DirectionVector(xa, ya, za, dxa, dya, dza)

    def airfoilToProfile(self):
        """Rotates from airfoil-aligned to profile

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the profile coordinate system

        """

        return DirectionVector(self.y, self.x, self.z, self.dy, self.dx, self.dz)

    def profileToAirfoil(self):
        """Rotates from profile to airfoil-aligned

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the airfoil-aligned coordinate system

        """

        return DirectionVector(self.y, self.x, self.z, self.dy, self.dx, self.dz)

    def cross(self, other):
        """cross product between two DirectionVectors

        Parameters
        ----------
        other : DirectionVector
            other vector to cross with

        Returns
        -------
        vector : DirectionVector
            vector = self X other

        """
        v1 = np.c_[self.x, self.y, self.z]
        v2 = np.c_[other.x, other.y, other.z]
        v = np.cross(v1, v2)

        if len(v.shape) > 1:
            return DirectionVector(v[:, 0], v[:, 1], v[:, 2])
        else:
            return DirectionVector(v[0], v[1], v[2])

    def cross_deriv(self, other, namea="a", nameb="b"):
        """defined only for floats for now"""

        # c = a X b
        a = self
        b = other

        dx = {}
        dx[namea] = np.r_[0.0, b.z, -b.y]
        dx[nameb] = np.r_[0.0, -a.z, a.y]

        dy = {}
        dy[namea] = np.r_[-b.z, 0.0, b.x]
        dy[nameb] = np.r_[a.z, 0.0, -a.x]

        dz = {}
        dz[namea] = np.r_[b.y, -b.x, 0.0]
        dz[nameb] = np.r_[-a.y, a.x, 0.0]

        return dx, dy, dz

    def cross_deriv_array(self, other, namea="a", nameb="b"):
        # c = a X b
        a = self
        b = other

        dx = {}
        dx["d" + namea + "x"] = np.zeros_like(b.x)
        dx["d" + namea + "y"] = b.z
        dx["d" + namea + "z"] = -b.y
        dx["d" + nameb + "x"] = np.zeros_like(a.x)
        dx["d" + nameb + "y"] = -a.z
        dx["d" + nameb + "z"] = a.y

        dy = {}
        dy["d" + namea + "x"] = -b.z
        dy["d" + namea + "y"] = np.zeros_like(b.y)
        dy["d" + namea + "z"] = b.x
        dy["d" + nameb + "x"] = a.z
        dy["d" + nameb + "y"] = np.zeros_like(a.y)
        dy["d" + nameb + "z"] = -a.x

        dz = {}
        dz["d" + namea + "x"] = b.y
        dz["d" + namea + "y"] = -b.x
        dz["d" + namea + "z"] = np.zeros_like(b.z)
        dz["d" + nameb + "x"] = -a.y
        dz["d" + nameb + "y"] = a.x
        dz["d" + nameb + "z"] = np.zeros_like(a.z)

        return dx, dy, dz

    def __neg__(self):
        """negate direction vector"""

        return DirectionVector(-self.x, -self.y, -self.z)

    def __add__(self, other):
        """add two DirectionVector objects (v1 = v2 + v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x + other.x, self.y + other.y, self.z + other.z)
        else:
            return DirectionVector(self.x + other, self.y + other, self.z + other)

    def __sub__(self, other):
        """subtract DirectionVector objects (v1 = v2 - v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x - other.x, self.y - other.y, self.z - other.z)
        else:
            return DirectionVector(self.x - other, self.y - other, self.z - other)

    def __iadd__(self, other):
        """add DirectionVector object to self (v1 += v2)"""

        if isinstance(other, DirectionVector):
            self.x += other.x
            self.y += other.y
            self.z += other.z
        else:
            self.x += other
            self.y += other
            self.z += other

        return self

    def __isub__(self, other):
        """subract DirectionVector object from self (v1 -= v2)"""

        if isinstance(other, DirectionVector):
            self.x -= other.x
            self.y -= other.y
            self.z -= other.z
        else:
            self.x -= other
            self.y -= other
            self.z -= other

        return self

    def __mul__(self, other):
        """multiply vector times a scalar or element by element muiltiply times another vector (v1 = alpha * v2 or v1 = v2 * v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x * other.x, self.y * other.y, self.z * other.z)
        else:
            return DirectionVector(self.x * other, self.y * other, self.z * other)

    def __truediv__(self, other):
        """divide vector by a scalar or element by element division with another vector (v1 = v2 / alpha or v1 = v2 / v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x / other.x, self.y / other.y, self.z / other.z)
        else:
            return DirectionVector(self.x / float(other), self.y / float(other), self.z / float(other))

    def __imul__(self, other):
        """multiply self times a scalar or element by element muiltiply times another vector (v1 *= alpha or v1 *= v2)"""

        if isinstance(other, DirectionVector):
            self.x *= other.x
            self.y *= other.y
            self.z *= other.z
        else:
            self.x *= other
            self.y *= other
            self.z *= other
        return self

    def __str__(self):
        """print string representation"""

        return "{0}, {1}, {2}".format(self.x, self.y, self.z)



"""The install package contains `InstallPhase` and its subclasses."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = ["Jake Nunemaker", "Rob Hammond"]
__email__ = ["jake.nunemaker@nrel.gov" "robert.hammond@nrel.gov"]

from .install_phase import InstallPhase  # isort:skip
from .oss_install import FloatingSubstationInstallation, OffshoreSubstationInstallation
from .cable_install import ArrayCableInstallation, ExportCableInstallation
from .jacket_install import JacketInstallation
from .mooring_install import MooringSystemInstallation
from .turbine_install import TurbineInstallation
from .monopile_install import MonopileInstallation
from .quayside_assembly_tow import MooredSubInstallation, GravityBasedInstallation
from .scour_protection_install import ScourProtectionInstallation



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

from .standard import TurbineInstallation
